# -*- coding: utf-8 -*-
# Copyright (c) 2013 - 2018, Gorka Zamora-López <gorka@Zamora-Lopez.xyz>
#
# Released under the Apache License, Version 2.0 (the "License");
# you may not use this software except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0

"""
FASTER SYNTHETIC NETWORK GENERATORS
===================================

This module duplicates some functions in the module 'gamodels.py'
and significantly accelerates their performance. This is achieved thanks to
the Numba package (http://numba.pydata.org) which compiles annotated Python
and NumPy code to LLVM on run-time. Therefore, the use of this module
requires Numba to be installed.

Due to the run-time approach of Numba, the original functions in 'galib.py'
will run faster in very small networks and Numba-based functions take the
lead as network size increases. The precise size limit varies from function
to function and I will report approximate valuesfor each of them. In general,
for any network of N > 100 nodes Numba-based functions run faster.

RANDOM NETWORK GENERATORS
-------------------------
RandomGraph_Numba
    Generates random graphs with N nodes and L links.


...moduleauthor:: Gorka Zamora-Lopez <galib@zamora-lopez.xyz>

"""
from __future__ import division, print_function, absolute_import

import numpy as np
import numpy.random
from numba import jit


############################################################################
"""RANDOM NETWORK GENERATORS"""
@jit
def RandomGraph_Numba(N, L, directed=False, selfloops=False):
    """Generates random graphs with N nodes and L links.

    Similar to an Erdos-Renyi (ER) graph with probability p = rho, where
    rho is the density of links. In ER graphs the total number of links
    varies in different realizations what is unsuitable to compare empirical
    networks with their random counterparts. RandomGraph() allows to create
    random graphs that always have the same number of links. The precise
    formula for rho depends on the options given. For an undirected graph
    with no self-loops allowed, rho = 1/2 * L / (N*(N-1)).

    Parameters
    ----------
    N : integer
        The size of the network (number of nodes).
    L : integer
        Number of links of the resulting random network.
    directed : Boolean
        True if a directed graph is desired. False, for an undirected graph.
    selfloops: Boolean
        True if self-loops are allowed, False otherwise.

    Returns
    -------
    adjmatrix : ndarray of rank-2, size NxN and dtype = int.
        The adjacency matrix of the generated random graph.

    Notes
    -----
    Make sure to specify the right number of links for the type of graph
    desired. Keep in mind that directed graphs have twice as many links
    (arcs) as graphs (edges).

    See Also
    --------
    ErdosRenyiGraph : Random graphs with given link probability.
    """
    # 0) SECURITY CHECKS. Make sure L is not too large.
    if directed:
        if selfloops:
            maxL = N**2
            if L > maxL:
                raise ValueError("L out of bounds, max(L) = N**2 = %d" %maxL)
        else:
            maxL = N*(N-1)
            if L > maxL:
                raise ValueError("L out of bounds, max(L) = N*(N-1) = %d" %maxL)
    else:
        if selfloops:
            maxL = 0.5*N*(N+1)
            if L > maxL:
                raise ValueError("L out of bounds, max(L) = 1/2*N*(N+1) = %d" %maxL)
        else:
            maxL = 0.5*N*(N-1)
            if L > maxL:
                raise ValueError("L out of bounds. For the options given, max(L) = 1/2*N*(N-1) = %d" %maxL)

    # 1) INITIATE THE MATRIX AND HELPERS
    adjmatrix = np.zeros((N,N), np.uint8)
    counter = 0

    # 2) GENERATE THE MATRIX
    while counter < L:
        # 2.1) Pick up two nodes at random
        source = int(N * numpy.random.rand())
        target = int(N * numpy.random.rand())

        # 2.2) Check if they can be linked, otherwise look for another pair
        if adjmatrix[source,target] == 1: continue
        if not selfloops and source == target: continue

        # 2.3) If the nodes are linkable, place the link
        adjmatrix[source,target] = 1
        if not directed:
            adjmatrix[target,source] = 1

        counter += 1

    return adjmatrix


############################################################################
"""NETWORK REWIRING ALGORITHMS"""


############################################################################
"""MODULAR AND HIERARCHICAL NETWORK MODELS"""


##



# -*- coding: utf-8 -*-
# Copyright (c) 2013 - 2018, Gorka Zamora-López <gorka@Zamora-Lopez.xyz>
#
# Released under the Apache License, Version 2.0 (the "License");
# you may not use this software except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
"""
FASTER GRAPH ANALYSIS DESCRIPTORS
=================================

This module duplicates the slowest functions of main module 'metrics.py'
and significantly accelerates their performance. This is achieved thanks to
the Numba package (http://numba.pydata.org) which compiles annotated Python
and NumPy code to LLVM on run-time. Therefore, the use of this module
requires Numba to be installed.

Due to the run-time approach of Numba, the original functions in 'metrics.py'
will run faster in very small networks and Numba-based functions take the
lead as network size increases. The precise size limit varies from function
to function and I will report approximate valuesfor each of them. In general,
for any network of N > 100 nodes Numba-based functions run faster.

BASIC CONNECTIVITY DESCRIPTORS
------------------------------
MatchingIndex_Numba
    Computes the number of common neighbours of every pair of nodes.

PATHS AND GRAPH DISTANCE FUNCTIONS
----------------------------------
FloydWarshall_Numba
    Computes the pathlength between all pairs of nodes in a network.


...moduleauthor:: Gorka Zamora-Lopez <galib@zamora-lopez.xyz>

"""
from __future__ import division, print_function, absolute_import

import numpy as np
from numba import jit
from . import metrics

# __all__ = ['MatchingIndex_Numba', 'FloydWarshall_Numba']


############################################################################
"""CONNECTIVITY AND DEGREE STATISTICS"""
@jit
def MatchingIndex_Numba(adjmatrix, normed=True):
    """Computes the number of common neighbours of every pair of nodes.

    The matching index of two nodes i and j is the number of common
    neighbours they are linked with.

    Returns same result as MatchingIndex() in metrics.py but uses an ndarray-based
    algorithm and the Numba package to accelerate the calculations. Recommended
    only for networks of size N > 250 nodes.

    Parameters
    ----------
    adjmatrix : ndarray of rank-2
        The adjacency matrix of the network.

    normed : Boolean, optional
        If 'normed=False', returns the number of common neighbours
        of two nodes i and j. If 'normed=True', the number of common
        neighbours is divided by the total number of different nodes they
        are linked with, so values range from 0 (no common neighbours)
        to 1 (all neighbours are common to i and j). Explicit links i-->j
        and j-->i are excluded because they do not contribute to the
        matching. For example:
            1 --> [2,3,4,5]
            2 --> [4,5,6]
            MI(1,2) = len([4,5])/len([3,4,5,6]) = 0.5

    Returns
    -------
    MImatrix : ndarray of rank-2 of ndtype 'float64'
        A matrix containing the matching index for all pairs of nodes.

    Notes
    -----
    - The function accepts weighted networks but it ignores the weights.
    - If adjmatrix is directed, calling MatchingIndex(adjmatrix) computes the
    matching of the output neighbours. Passing the transpose adjmatrix.T to
    the function computes the matching of the input neighbours.
    """
    N = len(adjmatrix)
    # Convert the adjacency matrix into a boolean matrix
    adjmatrix = adjmatrix.astype(np.bool)

    MImatrix = np.identity(N, np.float64)

    for i in range(N):
        for j in range(i,N):

            # Number of common neighbours (intersection of neighbourhoods)
            mi = (adjmatrix[i] * adjmatrix[j]).sum()

            if normed:
                # Size of the union of the neighbourhoods
                norm = (adjmatrix[i] + adjmatrix[j]).sum()
                # Avoid counting the explicit links i-->j and j-->i
                if adjmatrix[i,j]: norm -= 1
                if adjmatrix[j,i]: norm -= 1

                # Normalize and save the value avoiding ZeroDivision errors
                if norm > 0:
                    mi = np.float64(mi) / norm
                    MImatrix[i,j] = mi
                    MImatrix[j,i] = mi
            else:
                # Save the value
                mi = np.float64(mi)
                MImatrix[i,j] = mi
                MImatrix[j,i] = mi

    return MImatrix


###############################################################################
"""PATHS, CYCLES AND DISTANCE FUNCTIONS"""
def FloydWarshall_Numba(adjmatrix, weighted_dist=False):
    """Computes the pathlength between all pairs of nodes in a network.

    WARNING! This version returns the same output as 'FloydWarshall()'
        function in main metrics.py module but runs much faster (for networks
        of N > 100). It requires package Numba to be installed.

    Parameters
    ----------
    adjmatrix : ndarray of rank-2
        The adjacency matrix of the network.
    weighted_dist : boolean, optional
        If True, if the graph distances are computed considering the weights
        of the links. False, otherwise. If 'adjmatrix' is a weighted
        network but'weighted_dist = False', the weights of the links are
        ignored.

    Returns
    -------
    distmatrix : ndarray of rank-2
        The pairwise distance matrix dij of the shortest pathlength between
        nodes i and j.

    See Also
    --------
    FloydWarshall : Computes the pathlength between all pairs of nodes.
    """
    # 0) DEFINE THE CORE OF THE FW ALGORITHM, ACCELARATED BY 'Numba'
    @jit
    def FW_Undirected(distmatrix):
        """The Floyd-Warshall algorithm for undirected networks
        """
        N = len(distmatrix)
        for k in range(N):
            for i in range(N):
                for j in range(i,N):
                    d = distmatrix[i,k] + distmatrix[k,j]
                    if distmatrix[i,j] > d:
                        distmatrix[i,j] = d
                        distmatrix[j,i] = d

    @jit
    def FW_Directed(distmatrix):
        """The Floyd-Warshall algorithm for directed networks
        """
        N = len(distmatrix)
        for k in range(N):
            for i in range(N):
                for j in range(N):
                    d = distmatrix[i,k] + distmatrix[k,j]
                    if distmatrix[i,j] > d:
                        distmatrix[i,j] = d

    ########################################################################
    # 1) PREPARE FOR THE CALCULATIONS
    # 1.1) Initialize the distance matrix
    if weighted_dist:
        distmatrix = np.where(adjmatrix == 0, np.inf, adjmatrix)
    else:
        distmatrix = np.where(adjmatrix == 0, np.inf, 1)

    # 1.2) Find out whether the network is directed or undirected
    recip = metrics.Reciprocity(adjmatrix)

    # 2) RUN THE FLOYD-WARSHALL ALGORITHM USING FASTER FUNCTIONS (NUMBA)
    if recip==1.0:
        FW_Undirected(distmatrix)
    else:
        FW_Directed(distmatrix)

    return distmatrix


############################################################################
"""COMPONENTS, COMMUNITIES, K-CORES..."""


######################################################################
"""ROLES OF NODES IN NETWORKS WITH COMMUNITY (ASSORTATIVE) ORGANIZATION"""


#



# -*- coding: utf-8 -*-
# Copyright (c) 2013 - 2018, Gorka Zamora-López <gorka@Zamora-Lopez.xyz>
#
# Released under the Apache License, Version 2.0 (the "License");
# you may not use this software except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0

"""
SYNTHETIC NETWORK GENERATORS
============================

This module contains functions to generate typical synthetic networks,
including random networks and methods to rewire networks.

RANDOM NETWORK GENERATORS
-------------------------
Lattice1D
    Generates regular ring lattices (all nodes have same degree).
Lattice1D_FixLinks
    Generates ring lattices with desired number of links.
WattsStrogatzGraph
    Generates small-world networks as in the Watts & Strogatz model.
ErdosRenyiGraph
    Generates random graphs following the Erdos & Renyi model.
RandomGraph
    Generates random graphs with N nodes and L links.
BarabasiAlbertGraph
    Generates scale-free networks after the Barabasi & Albert model.
ScaleFreeGraph
    Generates scale-free graphs of given size and exponent.

NETWORK REWIRING/RANDOMIZATION ALGORITHMS
-----------------------------------------
RewireNetwork
    Randomises an input graph conserving the degrees of its nodes.
ModularityPreservingGraph
    Randomises an input graph conserving its modular structure.

HIERARCHICAL AND MODULAR (HM) NETWORK MODELS
--------------------------------------------
ModularHeterogeneousGraph
    Generates random modular networks of given module sizes and densities.
HMpartition
    Returns a partition of nodes for a hierarchical and modular network.
HMRandomGraph
    Generates random hierarchical and modular networks of desired number
    of hierarchical levels and modules.
HMCentralisedGraph
    Generates random hierarchical and modular networks of desired number
    of hierarchical levels and modules, with centralised inter-modular
    connectivity through local hubs.
RavaszBarabasiGraph
    Generates hierarchical networks following the Ravasz & Barabasi model.


...moduleauthor:: Gorka Zamora-Lopez <galib@zamora-lopez.xyz>

"""
from __future__ import division, print_function, absolute_import

import types
import numpy as np
import numpy.random
from .metrics import Reciprocity
from .tools import ExtractSubmatrix


############################################################################
"""RANDOM NETWORK GENERATORS"""
def Lattice1D(N,z):
    """Generates regular ring lattices.

    Each node is connected to its 2z closest neighbours (z on the left and z on
    the right). Hence, all nodes have same degree.

    Parameters
    ----------
    N : integer
        Size of the network (number of nodes).
    z : integer
        Diameter of the neighbourhood that every node connects with.
        For a given z, every node has degree 2*z in the resulting lattice.

    Returns
    -------
    adjmatrix : ndarray of rank-2 and integer type.
        The adjacency matrix of the 1-dimensional lattice.

    See Also
    --------
    Lattice1D_FixLinks : Generates ring lattices with desired number of links.
    """
    # 0) SECURITY CHECK
    if z > N//2:
        raise ValueError("Largest possible z = N/2 =", N//2 )

    if z == 0:
        return np.zeros((N,N), np.uint8)

    # 1) CREATE THE LATTICE
    adjmatrix = np.zeros((N,N), np.uint8)

    # 1.1) Create the first row according to the number of neighbours
    adjmatrix[0,1:z+1] = 1
    adjmatrix[0,-z:] = 1

    # 1.2) Use numpy.roll() to copy rotated version of the first row
    for i in range(1,N):
        adjmatrix[i] = np.roll(adjmatrix[0],i)

    return adjmatrix

def Lattice1D_FixLinks(N,L):
    """Generates a 1D ring lattice with given number of links.

    Because the total number of links L is fixed, the resulting ring is not a
    perfect regular lattice in which all nodes have exactly the same degree.
    The result is quasi-regular. The largest degree found will be kmax = <k> + 2
    and the smallest degree kmin = <k> - 2.

    Parameters
    ----------
    N : integer
        Size of the network (number of nodes).
    L : integer
        Number of links the resulting network must have.

    Returns
    -------
    adjmatrix : ndarray of rank-2 and integer type.
        The adjacency matrix of the 1-dimensional lattice.

    See Also
    --------
    Lattice1D : Generates regular ring lattices (all nodes have same degree)
    """
    # 0) SECURITY CHECK
    if L > int(0.5*N*(N-1)):
        raise ValueError( "Largest number of links 1/2 * N*(N-1) =", int(0.5*N*(N-1)) )

    if L == 0:
        return np.zeros((N,N), np.uint8)

    adjmatrix = np.zeros((N,N), np.uint8)

    # 1.2) Use numpy.roll() to copy rotated version of a pattern row
    counter = 0
    finished = False
    for k in range(1,N):
        if finished: break
        row = np.zeros(N,np.uint8)
        row[k] = 1

        for i in range(N):
            adjmatrix[i] += np.roll(row,i)
            counter += 1
            if counter == L:
                finished = True
                break

    return adjmatrix + adjmatrix.T

def WattsStrogatzGraph(N, z, prew, lattice=None):
    """Generates small-world networks as in the Watts & Strogatz model.

    See Watts & Strogatz, Nature 393(4) (1998) for details of the model.

    The algorithm rewires, clockwise, the links of the first neighbours,
    of the second neighbours, etc. with probability prew. If a link is
    chosen for rewiring, it can't create a self-loop nor a multiple edge.

    Parameters
    ----------
    N : integer
        Size of the network (number of nodes).
    z : integer
        Diameter of the neighbourhood that every node connects with.
        For a given z, every node has degree k = 2*z in the resulting lattice.
    prew : float, between 0 and 1.
        Probability that links of the 1D lattice to be rewired.
    lattice : ndarray of rank-2, optional.
        The adjacency matrix of a 1D lattice that has to be rewired.
        When several realizations of the model are desired, the initial
        lattice, usually the output of Lattice1D() function, can be passed
        what avoids having to create a new lattice for every realization and
        saving computational time. Recommended

    Returns
    -------
    adjmatrix : ndarray of rank-2 and integer type.
        The adjacency matrix of the rewired 1-dimensional lattice.

    Notes
    -----
    1) Unlike the Watts & Strogatz model in which k is the total number of
    neighbours of a node, here z represents the largest z-neighbourhood
    with which a node is connected (left and right). Hence, comparing the
    two parameters, k = 2*z.

    2) When several realizations of the model are desired, the initial
    lattice, usually the output of Lattice1D() function, can be passed
    to the optional argument 'lattice'. This avoids having to create a
    new lattice for every realization and saves computational time.
    Recommended for several realizations of large networks.

    3) The present algorithm is only valid for graphs. In order to obtain
    directed small-world networks, use the RewireNetwork() function on a
    previously created ring-lattice.

    See Also
    --------
    Lattice1D : Generates a 1D lattice ring network.
    RewiredNetwork : Randomized the links of a network with equal probability.
    """

    # 0) SECURITY CHECKS
    if (prew < 0 or prew > 1):
        raise ValueError( "Probability 'prew' out of bounds. Insert value between 0 and 1" )

    # 1) CREATE OR CHECK THE VALIDITY OF THE INITIAL RING-LATTICE
    # if type(lattice) == types.NoneType:
    if lattice == None:
        if z > N//2:
            raise ValueError("Largest possible z = N/2 =", N//2 )
        adjmatrix = Lattice1D(N,z)
    else:
        Nlatt = len(lattice)
        if N != Nlatt:
            raise ValueError( "N and size of given lattice od not match." )
        adjmatrix = lattice.copy()

    # 2) REWIRE THE LINKS CLOCKWISE AND BY RANK OF NEIGHBOURHOOD
    # For each rank of neighbourhood...
    for k in range(1,z+1):
        # Choose every node in clockwise direction
        for i in range(N):
            # 2.1) Rewire the link with neighbour i+k with probability prew
            if numpy.random.rand() <= prew:
                done = False
                while not done:
                    target = int( N * numpy.random.rand() )
                    # Avoid self-loops and double links
                    if target == i: continue
                    if adjmatrix[i,target]: continue

                    # Else, remove the old link and place the new one.
                    if i+k < N:
                        adjmatrix[i,i+k] = 0
                        adjmatrix[i+k,i] = 0
                    else:
                        adjmatrix[i,i+k-N] = 0
                        adjmatrix[i+k-N,i] = 0

                    adjmatrix[i,target] = 1
                    adjmatrix[target,i] = 1
                    done = True

    return adjmatrix

def ErdosRenyiGraph(N, p, directed=False, selfloops=False, outdtype=np.uint8):
    """Generates random graphs following the Erdos & Renyi model.

    In an Erdos-Renyi graph a link happens with probability p. Therefore,
    different relizations of the graph may have different number of links.
    In the ensemble average, the expected number of links of ER graphs
    is <L> = 1/2 * p * N**2 if self-loops are permitted, or
    <L> = 1/2 * p * N * (N-1) if no self-loops are permitted.

    Parameters
    ----------
    N : integer
        The size of the network (number of nodes).
    p : float between 0 and 1
        Probability of links.
    directed : Boolean
        True if a directed graph is desired, False if an undirected graph is
        desired.
    selfloops: Boolean
        True if self-loops are allowed, False otherwise.

    Returns
    -------
    adjmatrix : ndarray of rank-2 and size NxN
        The adjacency matrix of the generated random graph.

    See Also
    --------
    RandomGraph : Random graphs with specified number of links
    """
    # 0) SECURITY CHECKS. Make sure that adequate parameters are given.
    if (p < 0.0 or p > 1.0):
        raise ValueError( "Probability p out of bounds. Insert value between 0 and 1" )

    # 1) FOR DIRECTED GRAPHS
    # 1.1) Create a random matrix with normally distributed values
    adjmatrix = numpy.random.rand(N,N)
    # 1.2 Select the entries with value <= p
    adjmatrix = np.where(adjmatrix <= p, 1, 0).astype(outdtype)

    # 2) FOR UNDIRECTED GRAPHS
    if not directed:
        adjmatrix[np.tril_indices(N)] = 0
        adjmatrix += adjmatrix.T

    # 3) Remove the diagonal if no self-loops are desired
    if not selfloops:
        adjmatrix[np.diag_indices(N)] = 0

    return adjmatrix

def RandomGraph(N, L, directed=False, selfloops=False):
    """Generates random graphs with N nodes and L links.

    Similar to an Erdos-Renyi (ER) graph with probability p = rho, where
    rho is the density of links. In ER graphs the total number of links
    varies in different realizations what is unsuitable to compare empirical
    networks with their random counterparts. RandomGraph() allows to create
    random graphs that always have the same number of links. The precise
    formula for rho depends on the options given. For an undirected graph
    with no self-loops allowed, rho = 1/2 * L / (N*(N-1)).

    Parameters
    ----------
    N : integer
        The size of the network (number of nodes).
    L : integer
        Number of links of the resulting random network.
    directed : Boolean
        True if a directed graph is desired. False, for an undirected graph.
    selfloops: Boolean
        True if self-loops are allowed, False otherwise.

    Returns
    -------
    adjmatrix : ndarray of rank-2, size NxN and dtype = int.
        The adjacency matrix of the generated random graph.

    Notes
    -----
    Make sure to specify the right number of links for the type of graph
    desired. Keep in mind that directed graphs have twice as many links
    (arcs) as graphs (edges).

    See Also
    --------
    ErdosRenyiGraph : Random graphs with given link probability.
    """
    # 0) SECURITY CHECKS. Make sure L is not too large.
    if directed:
        if selfloops:
            maxL = N**2
            if L > maxL:
                raise ValueError( "L out of bounds, max(L) = N**2 =", maxL )
        else:
            maxL = N*(N-1)
            if L > maxL:
                raise ValueError( "L out of bounds, max(L) = N*(N-1) =", maxL )
    else:
        if selfloops:
            maxL = 0.5*N*(N+1)
            if L > maxL:
                raise ValueError( "L out of bounds, max(L) = 1/2*N*(N+1) =", maxL )
        else:
            maxL = 0.5*N*(N-1)
            if L > maxL:
                raise ValueError( "L out of bounds, max(L) = 1/2*N*(N-1) =", maxL )

    # 1) INITIATE THE MATRIX AND HELPERS
    adjmatrix = np.zeros((N,N), np.uint8)
    counter = 0

    # 2) GENERATE THE MATRIX
    while counter < L:
        # 2.1) Pick up two nodes at random
        source = int(N * numpy.random.rand())
        target = int(N * numpy.random.rand())

        # 2.2) Check if they can be linked, otherwise look for another pair
        if adjmatrix[source,target] == 1: continue
        if not selfloops and source == target: continue

        # 2.3) If the nodes are linkable, place the link
        adjmatrix[source,target] = 1
        if not directed:
            adjmatrix[target,source] = 1

        counter += 1

    return adjmatrix

def BarabasiAlbertGraph(N, m, outdtype=np.uint8):
    """Generates scale-free networks after the Barabasi & Albert model.

    The Barabasi and Albert model (Science 286 (1999)) creates networks with
    a scale-free degree distribution of exponent gamma = -3 by a growth
    process with preferential attachment. At each iteration a new node is
    included that connects to the existing nodes with probability proportional
    to their degree.
    In our implementation the network is initialized by a complete graph of
    size m + 1 nodes.

    Parameters
    ----------
    N : integer
        Number of nodes of the final network.
    m : integer
        Number of links that each new node makes during the growing process.
    outdtype : numpy compatible data type, optional.
        The data type of the resulting adjacency matrix of the scale-free
        network.

    Returns
    -------
    adjmatrix : ndarray of rank-2 and size NxN
        The adjacency matrix of the generated scale-free network.
    """

    # 1) INITIATE THE NETWORK AS A COMPLETE GRAPH OF SIZE m
    adjmatrix = np.zeros((N,N),outdtype)
    adjmatrix[:m+1,:m+1] = np.ones((m+1,m+1),outdtype)
    adjmatrix[np.diag_indices(m+1)] = 0

    # 2) PERFORM THE PREFERENTIAL ATTACHMENT GROWTH
    # 2.0) Create a list initially containing the hubs ki times
    nodelist = list(range(m+1))*m

    # 2.1) Add a new node
    for i in range(m+1,N):
        counter = 0
        neighbours = []
        while counter < m:
            # 2.2) Choose a node at random
            jid = int(len(nodelist)*numpy.random.rand())
            j = nodelist[jid]
            # Avoid double links
            if adjmatrix[i,j]: continue

            # 2.3) Connect the new node with the randomly chosen node
            adjmatrix[i,j] = 1
            adjmatrix[j,i] = 1
            neighbours.append(j)
            counter += 1

        # 2.4) Update nodelist for new probability in the next iteration
        nodelist += neighbours
        nodelist += m*[i]

    return adjmatrix

def ScaleFreeGraph(N, L, exponent=3.0, directed=False):
    """Generates scale-free graphs of given size and exponent.

    It follows the method proposed by Goh, Kahng & Kim 'Universal Behaviour
    of Load Distribution in SF networks' PRL 87 (2001). Every node is chosen
    with probability alpha = 1 / (gamma - 1) where gamma is the desired
    exponent.

    Parameters
    ----------
    N : integer
        Number of nodes of the final network.
    L : integer
        Number of links of the resulting random network.
    exponent : float (optional)
        The exponent (in positive) of the degree-distribution of the resulting
        networks. Recommended values between 2 and 3.
    directed : boolean, optional
        False if a graph is desired, True if digraphs are desired. In case
        of digraphs, both the input and the output degrees follow a scale-
        free distribution but uncorrelated between them.

    Returns
    -------
    adjmatrix : ndarray of rank-2 and size NxN.
        The adjacency matrix of the generated scale-free network.

    Notes
    -----
    In case of directed networks the input and the output degrees of the
    nodes are correlated, e.g., input hubs are also output hubs.
    """
    # 0) SECURITY CHECKS
    if directed:
        maxL = N*(N-1)
        if L > maxL:
            raise ValueError( "L out of bounds, max(L) = N*(N-1) =", maxL )
    else:
        maxL = 0.5*N*(N-1)
        if L > maxL:
            raise ValueError( "L out of bounds, max(L) = 1/2*N*(N-1) =", maxL )

    # 1) PREPARE FOR THE CALCULATIONS
    adjmatrix = np.zeros((N,N), np.uint8)

    # Create a degree sequence
    alpha = 1.0/(exponent - 1.0)
    nodeweights = (np.arange(N) +1)**-alpha
    # Probability of a node to be chosen
    nodeweights /= nodeweights.sum()
    nodecumprobability = nodeweights.cumsum()
    del nodeweights

    # 2) CONNECT THE NETWORK
    counter = 1
    while counter <= L:

        # 2.1) Choose two nodes to connect
        xhead = numpy.random.rand()     # A random number between 0 and 1
        xsum = np.sum(np.sign(nodecumprobability-xhead))
        head = int(0.5*(N-xsum))

        xtail = numpy.random.rand()
        xsum = np.sum(np.sign(nodecumprobability-xtail))
        tail = int(0.5*(N-xsum))

        # 2.2) Do not allow self loops and multiple edges
        if head == tail: continue
        if adjmatrix[head,tail]: continue

        # 2.3) If conditions are satisfied, make the link
        adjmatrix[head,tail] = 1
        if not directed:
            adjmatrix[tail,head] = 1
        counter += 1

    return adjmatrix


############################################################################
"""NETWORK REWIRING ALGORITHMS"""
def RewireNetwork(adjmatrix, prewire=10, directed=None, weighted=False):
    """Randomises an input graph conserving the degrees of its nodes.

    It uses the link switching method to rewire networks while conserving
    both the input and the output degrees of the nodes. See references:
    A.R. Rao & S. Bandyopadhyay, Sankhya, Ser. A 58, 225 (1996), and
    J.M. Roberts, Soc. Networks 22, 273 (2000).

    Parameters
    ----------
    adjmatrix : ndarray of rank-2
        The adjacency matrix of the network to be rewired. 'adjmatrix' itself
        won't be rewired but a new matrix is returned.
    prewire : float, optional
        Fraction of links to be rewired. See Usage for further instructions.
    directed : Boolean, optional
        Specifies the directedness of the returned network. If 'directed' is
        None (default), the function checks the directedness of the input
        network and performs the rewiring accordingly. To save computational
        time the parameter can be manually specified.
        If 'directed' is True, a directed network is returned regardless
        'adjmatrix' is directed or undirected. Input and output degrees are
        conserved.
        If 'directed' is False, an undirected network is returned. Only valid
        when 'adjmatrix' is undirected. DO NOT set 'directed = False' when
        'adjmatrix' is a directed network, degress won't be conserved and the
        function won't raise an error.
    weighted : Boolean, optional
        Specifies whether the weights of the links are conserved.
        If 'weighted' is False, a binary network is returned.
        If 'weighted' is True, links are switched conserving their weights.
        Notice, however, that it is impossible to conserve simultaneously
        both the degrees and the strength of all nodes. Hence, the weights of
        the links are conserved but only the input strength of the nodes
        are conserved. The output strength of nodes are not conserved.

    Returns
    -------
    rewmatrix : ndarray of rank-2
        The dtype of 'rewmatrix' depends on the option 'weighted'. If
        'weighted' is True, 'rewmatrix' is of same dtype as 'adjmatrix'.
        If 'weighted' is False, 'rewmatrix' has dtype uint8.

    Usage
    -----
    The parameter 'prewire' is not a probability, but a parameter that
    controls the number of iterations the link swithching procedure happens.
    At each iteration two links are switched, therefore for a given value of
    'prewire', prewire*L links are rewired in 1/2*prewire*L iterations, where
    L in the number of links. To make sure that all structure of the network
    has been completely randomized (appart from the degrees), values of
    prewire >= 5 are recommended.
    """
    # 0) PREPARE FOR THE CALCULATIONS
    # 0.1) Check the conditions for the rewiring process
    if directed==None:
        recip = Reciprocity(adjmatrix)
        if recip == 1.0: directed = False
        else: directed = True
    if weighted:
        rewmatrix = adjmatrix.copy()
    else:
        rewmatrix = np.where(adjmatrix,1,0).astype(np.uint8)

    N = len(rewmatrix)
    # 0.2) Generate the list of links
    if directed:
        linklist = np.array(rewmatrix.nonzero())
    else:
        # Apply nonzero only to the upper triangular part of the matrix
        linklist = np.array(np.triu(rewmatrix).nonzero())

    L = len(linklist[0])
    iterations = int(round(0.5*prewire*L))

    # DO THE REWIRING
    count = 0
    while count < iterations:
        # 1) SELECT TWO LINKS AT RANDOM:
        linkid1 = int(L*numpy.random.rand())
        linkid2 = int(L*numpy.random.rand())
        # Security check. If the two links are the same, discard the iteration
        if linkid1 == linkid2: continue

        h1 = linklist[0,linkid1]; t1 = linklist[1,linkid1]
        h2 = linklist[0,linkid2]; t2 = linklist[1,linkid2]

        # 2) SECURITY CHECKS TO AVOID INTRODUCING UNDESIRED LINKS
        # Avoid formation of self-loops
        if h1 == t2: continue
        if h2 == t1: continue
        # Avoid formation of double-links
        if rewmatrix[h1,t2]: continue
        if rewmatrix[h2,t1]: continue
        # Avoid trivial exchange of links
        if h1 == h2: continue
        if t1 == t2: continue

        # 3) IF ALL CONDITIONS SUCCESFUL, REWIRE
        # 3.1) Rewire the matrix
        if directed:
            # Put the new links
            if weighted:
                rewmatrix[h1,t2] = rewmatrix[h2,t2]
                rewmatrix[h2,t1] = rewmatrix[h1,t1]
            else:
                rewmatrix[h1,t2] = 1
                rewmatrix[h2,t1] = 1

            # Remove the old links
            rewmatrix[h1,t1] = 0
            rewmatrix[h2,t2] = 0

        else:
            # Put the new links
            if weighted:
                rewmatrix[h1,t2] = rewmatrix[h2,t2]
                rewmatrix[t2,h1] = rewmatrix[t1,h1]
                rewmatrix[h2,t1] = rewmatrix[h1,t1]
                rewmatrix[t1,h2] = rewmatrix[t2,h2]
            else:
                rewmatrix[h1,t2] = 1; rewmatrix[t2,h1] = 1
                rewmatrix[h2,t1] = 1; rewmatrix[t1,h2] = 1

            # Remove the old links
            rewmatrix[h1,t1] = 0.0; rewmatrix[t1,h1] = 0.0
            rewmatrix[h2,t2] = 0.0; rewmatrix[t2,h2] = 0.0

        # 3.2) Update the linklist
        linklist[1,linkid1] = t2
        linklist[1,linkid2] = t1

        # 3.3) Count the succesful realization
        count += 1

    return rewmatrix

def ModularityPreservingGraph(adjmatrix, partition, directed=None, selfloops=None):
    """Randomises an input graph conserving its modular structure.

    Given the adjacency matrix of a graph and a partition of its nodes, this
    function returns a network that contains the same number of links in
    each community and across them, but with the links randomly seeded.
    modules and across them as in the input graph.

    Parameters
    ----------
    adjmatrix : ndarray of rank-2
        The adjacency matrix of the network to be rewired. 'adjmatrix' itself
        won't be rewired but a new matrix is returned.
    partition : list of ndarrays of dtype = uint
        A list containing the indices of the nodes in each module.
    directed : Boolean (optional)
        True if a directed graph is desired, False if an undirected graph is
        desired.
    selfloops: Boolean (optional)
        True if self-loops are allowed, False otherwise.

    Returns
    -------
    adjmatrix : ndarray of rank-2 and dtype = uin8
        The adjacency matrix of the generated random modular graph.

    See Also
    --------
    HMRandomGraph : Generates Nested random hierarchical/modular networks.
    RewireNetwork : Randomises an input graph conserving degrees of nodes.
    ModularHeterogenousGraph : Generates random modular networks of given
                               module sizes and densities.
    """

    # 0) SECURITY CHECKS AND SETUP
    # Convert the input network into a boolean matrix
    adjmatrix = adjmatrix.astype('bool')

    # Check if the original network is directed or undirected
    if directed == None:
        if Reciprocity(adjmatrix) == 1: directed = False
        else: directed = True

    # Check if the original network accepts self-loops
    if selfloops == None:
        if adjmatrix.trace(): selfloops = True
        else: selfloops = False

    # 1) INITIATE THE NEW ADJACENCY MATRIX AND HELPERS
    N = len(adjmatrix)
    ncoms = len(partition)
    randmatrix = np.zeros((N,N), np.uint8)

    # 2) GENERATE THE BLOCK-WISE RANDOM GRAPH
    for c1, com1 in enumerate(partition):
        N1 = len(com1)

        # 2.1) Seed the random links within the module
        submatrix = ExtractSubmatrix(adjmatrix, com1)
        Lblock = submatrix.astype('bool').sum()
        if not directed:
            Lblock = int( round(np.float32(Lblock / 2)) )
        counter = 0
        while counter < Lblock:
            # Pick up two nodes at random
            idx1 = int( N1*numpy.random.rand() )
            idx2 = int( N1*numpy.random.rand() )
            # Check for self-loops
            if not selfloops and idx1==idx2: continue
            source = com1[idx1]
            target = com1[idx2]

            # Check if they are already connected
            if randmatrix[source,target] == 1: continue

            # If the nodes are linkable, place the link
            randmatrix[source,target] = 1
            if not directed:
                randmatrix[target,source] = 1

            counter += 1

        # 2.2) Seed the random links across modules
        for c2 in range(ncoms):
            if not directed and c2 < c1: continue
            if c1 == c2: continue

            com2 = partition[c2]
            N2 = len(com2)
            subnet = ExtractSubmatrix(adjmatrix,com1,com2)
            Lblock = subnet.astype('bool').sum()
            counter = 0
            while counter < Lblock:
                # Pick up two nodes at random, each from a different module
                idx1 = int( N1*numpy.random.rand() )
                idx2 = int( N2*numpy.random.rand() )
                source = com1[idx1]
                target = com2[idx2]

                # Check if they can be linked, otherwise look for another pair
                if randmatrix[source,target] == 1: continue

                # If the nodes are linkable, place the link
                randmatrix[source,target] = 1
                if not directed:
                    randmatrix[target,source] = 1

                counter += 1

    return randmatrix


############################################################################
"""MODULAR AND HIERARCHICAL NETWORK MODELS"""
def ModularHeterogeneousGraph(Nsizelist, pintlist, pext, directed=False, selfloops=False):
    """
    Generates random modular networks of given module sizes and densities.

    This function generates modular networks in which both the internal
    links within the modules and the external links across modules are shed
    at random. As in the Erdos-Renyi model, every pair of nodes is linked
    with the specified probability and therefore, the total number of links
    slightly varies from one realization to another. The user will specify
    the size and the internal connection probability for each of the modules,
    and one unique probability for the external links across communities.

    Parameters
    ----------
    Nsizelist : list, tuple or array of integers
        A list containing the desired size (number of nodes) for every
        module in the network.
    pintlist : list, tuple or array of floats
        A list  containing the internal link probability for each of the
        modules. All values must range between 0 and 1.
    pext : float
        The external probability of connection between nodes in different
        communities.
    directed : Boolean (optional)
        True if a directed graph is desired, False if an undirected graph is
        desired.
    selfloops: Boolean (optional)
        True if self-loops are allowed, False otherwise.

    Returns
    -------
    adjmatrix : ndarray of rank-2 and dtype = uin8
        The adjacency matrix of the generated random modular graph.
    partition : list of ndarrays of dtype = uint
        A list containing the indices of the nodes in each module.

    Usage and examples
    ------------------
    Setting Nsizelist = [100,200,300], pintlist = [0.3, 0.3, 0.5] and
    pext = 0.01 will generate a random graph of size N = 600 nodes with
    three modules of sizes N1 = 100, N2 = 200 and N3 = 300 nodes respectively.
    Each module is an Erdos-Renyi random graph with link probability
    (approximate density of links) pint1 = 0.3, pint2 = 0.3 and pint3 = 0.5
    respectively. The probability that a pair of nodes in different
    modules are connected is pext = 0.01.

    See Also
    --------
    HMRandomGraph : Generates Nested random hierarchical/modular networks.
    ErdosRenyiGraph : Generates random graphs with given link probability.
    """
    # 0) SECURITY CHECKS
    if len(Nsizelist) != len(pintlist):
        raise TypeError( "Parameters 'Nsizelist' and 'pintlist' not aligned." )
    if (pext < 0.0 or pext > 1.0):
        raise ValueError("Probability 'pext' out of bounds, insert value between 0 and 1" )
    for c in range(len(pintlist)):
        if (pintlist[c] < 0.0 or pintlist[c] > 1.0):
            raise ValueError( "Probability 'pintlist' out of bounds. Insert values between 0 and 1" )

    # 1) PREPARE TO CREATE THE NETWORK
    N = np.add.reduce(Nsizelist)
    ncommunities = len(Nsizelist)
    adjmatrix = np.zeros((N,N), np.uint8)

    # Define the partition
    counter = 0
    partition = []
    for c in range(ncommunities):
        com = np.arange(counter,counter+Nsizelist[c])
        partition.append(com)
        counter += Nsizelist[c]

    # 2) GENERATE THE RANDOM MODULAR NETWORK
    for c1, com1 in enumerate(partition):
        N1 = Nsizelist[c1]

        for c2, com2 in enumerate(partition):
            N2 = Nsizelist[c2]

            # 2.0) Choose the probability to use
            if c1 == c2: pthres = pintlist[c1]
            else: pthres = pext

            # 2.1) Create a random matrix with normally distributed values
            submatrix = numpy.random.rand(N1,N2)
            # 2.2) Select the entries with value <= p
            submatrix = np.where(submatrix <= pthres, 1, 0).astype(np.uint8)
            # 1.3) Copy the submatrix into the final adjmatrix
            imin = com1[0]; imax = com1[-1] + 1
            jmin = com2[0]; jmax = com2[-1] + 1
            adjmatrix[imin:imax,jmin:jmax] = submatrix

    # 3) IF GRAPH SHOULD BE UNDIRECTED...
    if not directed:
        adjmatrix[np.tril_indices(N, k=-1)] = 0
        adjmatrix += adjmatrix.T

    # 4) Remove the diagonal if no self-loops are desired
    if not selfloops:
        adjmatrix[np.diag_indices(N)] = 0

    return adjmatrix, partition

def HMpartition(HMshape):
    """Returns a partition of nodes for a hierarchical and modular network.
    network.

    Parameters
    ----------
    HMshape : list, tuple or array-like.
        HMshape must be a list of integer numbers indicating the target
        hierarchical and modular structure. For example, HMshape = [2,4,20]
        will create a network of 2 modules, each module is divided into 4
        modules of 20 nodes each.

    Returns
    -------
    partitions : list
        Contains the partition matrices for each of the nlevel-1 hierarchical
        levels of a network created with given parameter 'HMshape'.

    See Also
    --------
    HMRNetwork : Generates random modular and hierarchical networks.
    """
    N = np.multiply.reduce(HMshape)
    nlevels = len(HMshape)

    partitions = []
    for level in range(nlevels):
        # 2.1.1) Find the number of blocks per hierarchical level
        if level == 0: continue

        nblocks = int(np.multiply.reduce(HMshape[:level]))
        partitionmatrix = np.zeros((N,nblocks), np.uint8)

        # 2.2) For each block in the hierarchical level
        for b in range(nblocks):
            # 2.2.1) Find the indices for the block
            Nblock = N // nblocks
            iminblock = b*Nblock
            imaxblock = (b+1)*Nblock

            partitionmatrix[iminblock:imaxblock,b] = 1

        partitions.append(partitionmatrix)

    return partitions

def HMRandomGraph(HMshape, avklist, directed=False, outdtype=np.uint8):
    """Generates random hierarchical and modular networks of desired number
    of hierarchical levels and modules.

    This function generalizes the benchmark hierarchical and modular network
    models introduced in [M.E.J. Newman & M. Girvan, Phys. Rev. E 69, 026113
    (2004)] and in [A. Arenas, A. Diaz-Guilera & C.J. Perez-Vicente, Phys.
    Rev. Lett. 96, 114102 (2006)]. It creates networks of networks that are
    randomly connected at each hierarchical level, i.e., random networks
    that are randomly connected with each other forming larger modules that
    randomly connect with each other. The communities of each level are
    of same size and link density.

    Parameters
    ----------
    HMshape : list, tuple or array-like.
        HMshape must be a list of integer numbers indicating the target
        hierarchical and modular structure. For example, HMshape = [2,4,20]
        will create a network of 2 modules, each module is divided into 4
        modules of 20 nodes each.
    avklist : list, tuple or array-like.
        The mean degree of the nodes at each hierarchical level. For example,
        avklist = [1,3,20] will generate a network in which nodes will have,
        on average, 20 links with their neighbours at the lowest level
        submodule they belong to, 3 links with the rest of neighbours at the
        submodule they belong to at the second level, and 1 link with any
        node at any community of the rest of the network.
    directed : boolean, optional.
        If true, the resulting network will be directed, else, it will be
        undirected.
    outdtype : data type, optional
        Data-type of the resulting adjacency matrix. Default: uint8.

    Returns
    -------
    adjmatrix : ndarray of rank-2.
        The adjacency matrix of the generated modular and hierarchical
        network.

    Usage and examples
    ------------------
    The function accepts any desired number of hierarchical levels and
    partitions of the levels into desired number of nodes or communities.
    - HMshape = [4,4,16] will create a network of 4x4x16 = 256 nodes and
    three hierarchical levels. The network contains 4 communities, each
    subdivided into 4 communities of 16 nodes.
    - HMshape = [2,5,3,20] will generate a network of 2x5x3x20 = 600 nodes
    and four hierarchical levels. The network is composed of 2 communities,
    each subdivided into 5 communities, each subdivided into 3 communities
    of 20 nodes each.

    The parameter 'avklist' controls the density of connections between
    the modules in each of the hierarchical levels.
    - avklist = [1,3,13], for the first example above, means that each node
    is connected with 13 nodes of the first level, to 3 nodes in other
    communities at the second level and makes one link with nodes in the
    rest of the network which do not belong to the same modules in either
    the first or the second level.

    In order to generate the 13-4 or the 15-2 networks in [Arenas et al,
    Phys. Rev. Lett., 114102 (2006)], the function has to be called with
    parameters HMshape = [4,4,16] and avklist = [1,4,13] or avklist = [1,2,15]
    respectively.

    Notes
    -----
    The degrees given with parameter 'avklist' are average degrees. Due to
    the random generation process, every node will only have the desired
    internal and external degrees on the ensemble average, but not
    necessarily in each realization.

    See Also
    --------
    HMCentralisedGraph
        Random hierarchical and modular network with inter-modular
        connectivity centralised through local hubs.
    ModularHeterogeneousGraph
        Generates random modular networks with desired module sizes and
        densities.
    """
    def SeedLinks(adjmatrix, L, partition, directed=False):
        """This is a helper subfunction that seeds links at random.
        """
        # 0) SSECURITY CHECK
        if len(partition) < 2.0:
            raise TypeError( "Partition needs to have at least 2 communities" )

        ncoms = len(partition)
        Ncoms = len(partition[0])
        counter = 0
        while counter < L:
            # 1) Choose two communities at random
            com1idx = int(ncoms*numpy.random.rand())
            com2idx = int(ncoms*numpy.random.rand())
            # Make sure the nodes belong to different communities
            if com1idx == com2idx: continue

            # 2) Choose one node from each community
            nodeidx1 = int( Ncoms*numpy.random.rand() )
            nodeidx2 = int( Ncoms*numpy.random.rand() )
            node1 = partition[com1idx][nodeidx1]
            node2 = partition[com2idx][nodeidx2]
            # Avoid multiple links
            if adjmatrix[node1,node2]: continue

            # Link the nodes and update counter
            if directed:
                adjmatrix[node1,node2] = 1
                counter += 1
            else:
                adjmatrix[node1,node2] = 1
                adjmatrix[node2,node1] = 1
                counter += 1

    #______________________________________________________________________
    # 0) SECURITY CHECKS
    if len(HMshape) != len(avklist):
        raise ValueError( "HMshape and plist not aligned." )
    if HMshape[0] < 1:
        raise ValueError("HMshape[0] <= 1. First hierarchical level must contain more than one module." )

    # 1) PREPARE TO CREATE THE NETWORK
    N = np.multiply.reduce(HMshape)
    nlevels = len(HMshape)
    adjmatrix = np.zeros((N,N), outdtype)

    # 2) CREATE THE HM NETWORK BY SEEDING LINKS AT DIFFERENT SCALES
    # 2.1) For each hierarchical level
    for level in range(nlevels):
        # 2.1.1) Find the number of blocks per hierarchical level
        if HMshape[level] == 1: continue

        if level == 0: nblocks = 1
        else: nblocks = int(np.multiply.reduce(HMshape[:level]))

        # 2.2) For each block in the hierarchical level
        for b in range(nblocks):
            # 2.2.1) Find the indices for the block
            Nblock = N // nblocks
            iminblock = b*Nblock
            imaxblock = (b+1)*Nblock

            # 2.2.2) Create the partition of nodes into communities within
            # the current block in the current hierarchical level
            ncoms = HMshape[level]
            Ns = Nblock // ncoms
            partition = []
            for p in range(ncoms):
                partition.append(range(iminblock + p*Ns,iminblock + (p+1)*Ns))

            # 3) SEED THE LINKS AT RANDOM BETWEEN COMMUNITIES
            # NOTE: In the last hierarchical level, every node is one community
            if directed:
                Ls = Nblock * avklist[level]
            else:
                Ls = 0.5 * Nblock * avklist[level]

            SeedLinks(adjmatrix, Ls, partition, directed)

    return adjmatrix

def HMCentralisedGraph(HMshape, avklist, gammalist, directed=False, outdtype=np.uint8):
    """    Generates random hierarchical and modular networks of desired number
    of hierarchical levels and modules, with centralised inter-modular
    connectivity through local hubs.

    This function creates networks of networks that are randomly connected
    at each hierarchical level. The submodules at the lowest level are
    random scale-free-like graphs. Setting a high exponent for this
    level (e.g. gamma=[x,x,100]) will make them usual random graphs.
    At each heirarchical level, the links between modules are seeded
    following a preferential attachment rule (same as as for the generation
    of scale-free graphs). The nodes at every module are assigned different
    probability to link with other modules, hence, the inter-modular links
    end up concentrated on few hubs, with every community having its own
    set of hubs.

    For further details see reference:
    - G. Zamora-Lopez, Y. Chen et al. "Functional complexity emerging from
    anatomical constraints in the brain: the significance of network
    modularity and rich-clubs." Scientific Reports 6:38424 (2016).

    Parameters
    ----------
    HMshape : list, tuple or array-like.
        HMshape must be a list of integer numbers indicating the target
        hierarchical and modular structure. For example, HMshape = [4,50]
        will create a network composed of 4 modules of 50 nodes each.
        HMshape = [2,4,50] will create a network of 2 modules, each module
        divided into 4 submodules of 50 nodes each.
    avklist : list, tuple or array-like.
        The mean degree of the nodes at each hierarchical level. For example,
        avklist = [1,3,20] will generate a network in which nodes will have,
        on average, 20 links with their neighbours at the lowest level
        submodule they belong to, 3 links with the rest of neighbours at the
        submodule they belong to at the second level, and 1 link with any
        node at any community of the rest of the network.
    gammalist : list, tuple or array-like.
        Exponent controling the preferential attachement rule at each
        level. For a network with HMshape = [4,50], setting gammalist =
        [2.0,3.0] will create a network of four modules of 50 nodes each.
        Internally, the four submodules are scale-free-like networks with
        exponent gamma = 3.0. The links between modules are seeded at random
        but
    directed : boolean, optional.
        If true, the resulting network will be directed, else, it will be
        undirected.
    outdtype : data type, optional
        Data-type of the resulting adjacency matrix. Default: uint8.

    Returns
    -------
    adjmatrix : ndarray of rank-2.
        The adjacency matrix of the generated modular and hierarchical
        network.

    Usage and examples
    ------------------
    The function accepts any desired number of hierarchical levels and
    partitions of the levels into desired number of nodes or communities.
    - HMshape = [4,4,16] will create a network of 4x4x16 = 256 nodes and
    three hierarchical levels. The network contains 4 communities, each
    subdivided into 4 communities of 16 nodes.
    - HMshape = [2,5,3,20] will generate a network of 2x5x3x20 = 600 nodes
    and four hierarchical levels. The network is composed of 2 communities,
    each subdivided into 5 communities, each subdivided into 3 communities
    of 20 nodes each.

    The parameter 'avklist' controls the density of connections between
    the modules in each of the hierarchical levels.
    - avklist = [1,3,13], for the first example above, means that each node
    is connected with 13 nodes of the first level, to 3 nodes in other
    communities at the second level and makes one link with nodes in the
    rest of the network which do not belong to the same modules in either
    the first or the second level.

    The parameter 'gammalist' controls for the probability that every node
    has to be selected while links are seeded. This probability may change
    at different hierarchical levels.
    - gammalist = [2,100] with HMshape = [4,50], will lead to a modular
    network consisting of four random graphs (random because of the high
    gamma=100 at the level of the modules). To seed the inter-modular
    links, two nodes are chosen at random from two different communities.
    Internally, the probability of a node to be chosen differs such that
    each community will have its own hubs such that intermodular links
    preferentially occur between those hubs. The probability of a node within
    a module to be chosen as the target of an intermodular link is
    determined by the exponent gamma = 2, which would eventually to a
    schale-free graph of gamma = 2 in the limit of large networks.

    See Also
    --------
    HMRandomGraph
        Generates random hierarchical/modular network of given shape.
    ModularHeterogeneousGraph
        Generates random modular networks with desired module sizes and
        densities.
    """
    def GenerateBlock(L, partition, cumprobability, directed=False):
        """This is a helper function which generates the modules at a given
        hierachical level.
        """
        if len(partition) < 1:
            raise ValueError( "No modules given, Generating a random graph" )

        ncoms = len(partition)
        Ns = len(partition[0])
        N = ncoms*Ns

        blockmatrix = np.zeros((N,N),np.uint8)
        counter = 0
        while counter < L:
            # 1) Choose two communities at random
            com1idx = int(ncoms*numpy.random.rand())
            com2idx = int(ncoms*numpy.random.rand())
            # Make sure the communities are different
            if com1idx == com2idx: continue

            # 2) Choose two nodes to connect given the cumulative probabilities
            x = numpy.random.rand()     # A random number between 0 and 1
            xsum = np.sum(np.sign(cumprobability-x))
            idx = int(0.5*(Ns-xsum))
            head = com1idx * Ns + idx

            x = numpy.random.rand()
            xsum = np.sum(np.sign(cumprobability-x))
            idx = int(0.5*(Ns-xsum))
            tail = com2idx * Ns + idx

            # 3.2) Do not allow self loops and multiple edges
            if head == tail: continue
            if blockmatrix[head,tail]: continue

            # 3.3) If conditions are satisfied, make the link
            blockmatrix[head,tail] = 1
            if not directed:
                blockmatrix[tail,head] = 1
            counter += 1

        return blockmatrix

    def SkewedRandomGraph(L, cumprobability, directed=False):
        """This is a helper function to seed inter-modular links.
        """
        Ncom = len(cumprobability)

        blockmatrix = np.zeros((Ncom,Ncom), np.uint8)
        counter = 0
        while counter < L:
            # 1) Choose two nodes to connect given the cumulative probabilities
            x = numpy.random.rand()
            xsum = np.sum(np.sign(cumprobability-x))
            head = int(0.5*(Ncom-xsum))

            x = numpy.random.rand()
            xsum = np.sum(np.sign(cumprobability-x))
            tail = int(0.5*(Ncom-xsum))

            # 3.2) Do not allow self loops and multiple edges
            if head == tail: continue
            if blockmatrix[head,tail]: continue

            # 3.3) If conditions are satisfied, make the link
            blockmatrix[head,tail] = 1
            if not directed:
                blockmatrix[tail,head] = 1
            counter += 1

        return blockmatrix

    #______________________________________________________________________
    # 0) SECURITY CHECKS
    if len(HMshape) != len(avklist):
        raise ValueError( "HMshape and avklist are not alighned." )
    if gammalist:
        if len(gammalist) != len(HMshape):
            raise ValueError( "HMshape and gammalist are not alighned." )

    # 1) PREPARE TO CREATE THE NETWORK
    N = np.multiply.reduce(HMshape)
    nlevels = len(HMshape)
    adjmatrix = np.zeros((N,N), outdtype)

    # 1.1) If no hub parameters given, connect modules at random.
    # This case returns the same networks as 'HMRandomGraph' function.
    if not gammalist:
        alpha = np.ones(nlevels, float)
    else:
        alpha = 1.0 / (np.array(gammalist,float) - 1.0)

    # 2) CREATE THE HM NETWORK BY SEEDING LINKS AT DIFFERENT SCALES
    for level in range(nlevels-1):
        # 2.1) Find the number of blocks, communities and nodes
        # Number of blocks
        if HMshape[level] == 1: continue
        if level == 0: nblocks = 1
        else:
            nblocks = int(np.multiply.reduce(HMshape[:level]))
        # Number of nodes per block
        Nblock = np.multiply.reduce(HMshape[level:])
        # Number of communities per block
        ncoms = HMshape[level]
        # Number of nodes per community
        Ncom = Nblock // ncoms

        # 2.2) Define typical partition for the current hierarchical level
        partition = np.zeros((ncoms,Ncom), np.uint)
        for i in range(ncoms):
            partition[i] = np.arange(i*Ncom,(i+1)*Ncom, dtype=np.uint)

        # 2.3) Define the typical node selection probabilities within a block
        if level < nlevels - 2:
            nodeweights = np.ones(Ncom,np.float)
            ncomsnext = HMshape[level+1]
            Ncomnext = Ncom // ncomsnext
            for i in range(ncomsnext):
                nodeweights[i*Ncomnext:(i*Ncomnext+Ncomnext)] = \
                    ((np.arange(Ncomnext)+1).astype(float))**-alpha[level]
            # Probability of a node to be chosen
            nodeweights /= nodeweights.sum()
            cumprobability = nodeweights.cumsum()
        else:
            nodeweights = np.ones(Ncom,np.float)
            nodeweights = ((np.arange(Ncom)+1).astype(float))**-alpha[level]
            # Probability of a node to be chosen
            nodeweights /= nodeweights.sum()
            cumprobability = nodeweights.cumsum()

        # 2.4) Determine the number of links to be seed per block
        if directed:
            Lblock = Nblock * avklist[level]
        else:
            Lblock = 0.5 * Nblock * avklist[level]

        # 2.5) Seed intercommunity links in the current hierarchical level
        for b in range(nblocks):
            minidx = b*Nblock
            maxidx = (b+1)*Nblock
            adjmatrix[minidx:maxidx,minidx:maxidx] = \
                GenerateBlock(Lblock, partition, cumprobability, directed)

    # 3) CREATE THE LAST HIERARCHICAL LEVEL
    # 3.1 ) Find the number communities and nodes per community
    Ncom = HMshape[-1]
    ncoms = len(adjmatrix) // Ncom

    # 3.2) Define the typical node selection probabilities within a community
    nodeweights = ((np.arange(Ncom)+1).astype(np.float))**(-alpha[-1])
    nodeweights /= nodeweights.sum()    # Probability of a node to be chosen
    cumprobability = nodeweights.cumsum()

    # 3.3) Determine the number of links to be seed per community
    if directed:
        Lcom = Ncom* avklist[-1]
    else:
        Lcom = 0.5 * Ncom * avklist[-1]

    # 3.4) Seed intercommunity links in the current hierarchical level
    for i in range(ncoms):
        minidx = i*Ncom
        maxidx = (i+1)*Ncom

        adjmatrix[minidx:maxidx,minidx:maxidx] = \
            SkewedRandomGraph(Lcom, cumprobability, directed)

    return adjmatrix

def RavaszBarabasiGraph(Nmotif=4, hlevels=3, hublinks=True):
    """Generates a Ravasz-Barabasi hierarchical graph.

    The Ravasz-Barabasi graph is a pseudo-fractal network that starts from a
    basic motif, a complete graph of size 'Nmotif'. Further hierarchical levels
    are constructed replicating the original motif, by making each of the nodes
    in original motif as the central node of the new replicated motif. All new
    'peripheral' vertices link

    The network is claimed to produce scale-free-like degree distribution but
    modular structure is rather unclear. Two different versions of the model
    where published. In Ref. [science2002], the hubs of each hierarchical level
    are connected forming a ring graph. Note, however, that these links
    between the hubs do not lead to a rich-club. The model in Ref. [pre2003],
    ignored the connections among the local hubs at same hierarchical level hubs.
    This function creates both versions, controled by parameter 'hublinks'.

    Parameters
    ----------
    Nmotif : integer
        Size of the basic motif.
    hlevels : integer
        Number of hierarchical levels. Notice that hlevels = 1 will return the
        basic motif, a complete graph of size Nmotif.

    Returns
    -------
    adjmatrix : ndarray of rank-2.
        The adjacency matrix of the generated scale-free network. The size of
        the final network is N = Nmotif^hlevels.

    References
    ----------
    [science2002] E. Ravasz, A.L. Somera, D.A. Mongru, Z.N. Oltvai and A.L.
    Barabasi "Hierarchical organization of modularity in metabolic networks."
    Science 297, 1551-1555 (2002).
    [pre2003] E. Ravasz and A. Barabasi "Hierarchical organization in complex
    networks" Phys. Rev. E 67, 026112 (2003).
    """
    # 0) SECURITY CHECKS
    if Nmotif < 4:
        raise ValueError('Basic motif needs at least four nodes, Nmotif > 3.')
    if hlevels < 1:
        raise ValueError('Number of hirarchical levels needs to be positive, hlevels > 0.')

    # The number of nodes of one module in three different scales
    N = Nmotif**(hlevels)
    adjmatrix = np.zeros((N,N), np.uint8)

    # 1) DEFINE THE BASIC MOTIF
    adjmatrix[:Nmotif,:Nmotif] = 1
    adjmatrix[np.diag_indices(Nmotif)] = 0

    # 2) GROW AND CONNECT THE NETWORK
    for lev in range(1,hlevels):
        hublist = np.arange(Nmotif**lev,Nmotif**(lev+1),Nmotif**lev)

        # 2.1) Link the hubs at the new level forming a ring
        if hublinks:
            for i in range(Nmotif-2):
                node1 = hublist[i]
                node2 = hublist[i+1]
                adjmatrix[node1,node2] = 1
                adjmatrix[node2,node1] = 1
            adjmatrix[hublist[0],hublist[-1]] = 1
            adjmatrix[hublist[-1],hublist[0]] = 1

        # 2.2) Define the basis motif for the current level
        motif = adjmatrix[:Nmotif**lev,:Nmotif**lev]
        newNmotif = len(motif)

        # 2.3) Make copies of the motif at each hub location
        for hub in hublist:
            adjmatrix[hub:hub+newNmotif,hub:hub+newNmotif] = motif

        # 2.4) Connect all new peripheral nodes with the central hub
        for i in range(Nmotif**lev,Nmotif**(lev+1)):
            if i not in hublist:
                adjmatrix[0,i] = 1
                adjmatrix[i,0] = 1

    return adjmatrix

#



# -*- coding: utf-8 -*-
# Copyright (c) 2013 - 2018, Gorka Zamora-López <gorka@Zamora-Lopez.xyz>
#
# Released under the Apache License, Version 2.0 (the "License");
# you may not use this software except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0

"""
MISCELLANOUS HELPER FUNCTIONS
=============================

This module contains miscellaneous helper functions useful for the analisys
of graphs and complex networks.

I/O AND DATA CONVERSIONS
------------------------
LoadFromPajek
    Reads a network from a text file with Pajek format.
Save2Pajek
    Saves a network into a text file with Pajek format.
LoadLabels
    Reads the labels of nodes from a text file.
SaveLabels
    Saves the labels of nodes into a text file
ReadPartition
    Reads a partition of nodes from a text file.
SavePartition
    Saves a partition of nodes into a text file.
ExtractSubmatrix
    Returns the sub-matrix composed by a set of nodes.
SymmetriseMatrix
    Converts a directed network into undirected by averaging the weights.
LaplacianMatrix
    Returns the Laplacian matrix of a given network.
CleanPaths
    Finds and removes in-place repeated, opposite paths from a list of paths.

ARRAY AND MATRIX COMPARISONS
----------------------------
ArrayCompare
    Compares whether two arrays are identical or not.
HammingDistance
    Computes the Hamming distance between two arrays of same shape.

ADDITIONAL MATH FUNCTIONS
-------------------------
NonZeroMin
    Returns the smallest non-zero value in an array.
CumulativeDistribution
    Computes the cumulative distribution of a dataset.
Factorial
    Computes the factorial of an integer number.
BinomialCoefficient
    Computes the binomial coefficient of n over m.
Quartiles
    Finds the Q1, Q2 and Q3 quartiles of a dataset.
AllPermutations
    Given a set, it returns all possible permutations.
AllCominations
    Given a set, finds all combinations of given size.
AllBipartitions
    Given a set, finds all its possible bipartitions.
MeanCorrelation
    Computes the Fisher-corrected mean value of correlation values.


...moduleauthor:: Gorka Zamora-Lopez <galib@zamora-lopez.xyz>

"""
from __future__ import division, print_function, absolute_import

import itertools
import types
import re
import warnings
import functools

import numpy as np

# from . import metrics
import galib.metrics


# I/O AND DATA CONVERSIONS ################################################
def LoadFromPajek(filepath, getlabels=False):
    """Reads a network from a text file with Pajek format.

    Parameters
    ----------
    filepath : string
        The source .net file of the network in Pajek format.
    getlabels : boolean, optional
        If True, the function also reads and returns the labels of the nodes.

    Returns
    -------
    adjmatrix : ndarray of rank-2
        The adjacency matrix of the network.
    labels : list, optional
        If getlabels = True, the function also return a list containing the
        names of the nodes.

    See Also
    --------
    LoadLabels, SaveLabels
    Save2Pajek : Saves a network into a file in Pajek-readable format.

    Notes
    -----
    1. The function automatically determines whether the network is directed,
    undirected and / or weighted.
    2. The returned adjacency matrix is of dtype 'int' or 'float', depending
    on the weights in the file.
    """
    # 0) OPEN THE FILE AND READ THE SIZE OF THE NETWORK
    pajekfile = open(filepath, 'r')
    firstline = pajekfile.readline()
    firstline = firstline.split()
    N = int(firstline[1])

    # 1) READ THE LABELS OF THE NODES IF WANTED
    if getlabels:
        labels = []

        # Security check, make sure that labels of nodes are listed in file
        line = pajekfile.readline()
        if line.split()[0] != '1':
            pajekfile.seek(1)
            print('LoadFromPajek() warning: No labels found to read.')

        # If labels are in file continue reading the labels.
        else:
            # If labels are wrapped in between quotes
            try:
                idx1 = line.index('"') + 1
                # Add the first label
                idx2 = line[idx1:].index('"')
                label = line[idx1:idx1+idx2]
                labels.append(label)

                # And now read the labels for the rest of the nodes
                for i in range(1,N):
                    line = pajekfile.readline()
                    idx1 = line.index('"') + 1
                    idx2 = line[idx1:].index('"')
                    label = line[idx1:idx1+idx2]
                    labels.append(label)

            # Otherwise, make a wild guess of what the label is
            except ValueError:
                # Add the first label
                label = line.split()[1]
                labels.append(label)

                # And now read the labels of the rest of the nodes
                for i in range(1,N):
                    line = pajekfile.readline()
                    label = line.split()[1]
                    labels.append(label)

    # 2) READ THE LINKS AND CREATE THE ADJACENCY MATRIX
    # 2.1) Find out whether the network is directed or undirected
    # while loop to skip empty lines if needed or the lines of the labels
    done = False
    while not done:
        line = pajekfile.readline()
        if line[0] == '*':
            if 'Edges' in line:
                directed = False
            elif 'Arcs' in line:
                directed = True
            else:
                print('Could not find whether network is directed or undirected')
                break
            done = True

    # 2.2) Read the first line contining a link
    line = pajekfile.readline()
    line = line.split()

    # If link information is BINARY, just read the adjacency list links
    if len(line) == 2:
        # 2.3) Declare the adjacency matrix and include the first link
        adjmatrix = np.zeros((N,N), np.uint8)
        i = int(line[0]) - 1
        j = int(line[1]) - 1
        adjmatrix[i,j] = 1
        if not directed:
            adjmatrix[j,i] = 1

        # 2.4) Include the rest of the links
        for line in pajekfile:
            i, j = line.split()
            i = int(i) - 1
            j = int(j) - 1
            adjmatrix[i, j] = 1
            if not directed:
                adjmatrix[j, i] = 1

    # If the link information is WEIGHTED, read the weighted links
    elif len(line) == 3:
        # 2.3) Find whether link weights are integer or floating poing
        i, j, aij = line
        outdtype = np.int
        try:
            outdtype(aij)
        except ValueError:
            outdtype = np.float

        # 2.4) Declare the adjacency matrix and include the first link
        adjmatrix = np.zeros((N, N), outdtype)
        i = int(i) - 1
        j = int(j) - 1
        adjmatrix[i, j] = outdtype(aij)
        if not directed:
            adjmatrix[j, i] = outdtype(aij)

        # 2.5) Read the rest of the file and fill-in the adjacency matrix
        for line in pajekfile:
            i, j, aij = line.split()
            i = int(i) - 1
            j = int(j) - 1
            adjmatrix[i, j] = outdtype(aij)
            if not directed:
                adjmatrix[j, i] = adjmatrix[i, j]

    # 3) CLOSE FILE AND RETURN RESULTS
    pajekfile.close()

    if getlabels:
        return adjmatrix, labels
    else:
        return adjmatrix

def Save2Pajek(filepath, adjmatrix, labels=[], directed=False, weighted=False):
    """Saves a network into a text file with Pajek format.

    Parameters
    ----------
    adjmatrix : ndarray of rank-2
        The adjacency matrix of the network.
    filepath : string
        The target file in which the partition is to be saved.
    labels : list, optional
        A list containing the labels of each node.
    directed : Boolean, optional
        True if the network is directed, False otherwise.
    weighted : Boolean, optional
        If 'True', the weights of the links will be included in the file.
        Otherwise, it assumes the network is binary and it saves only the
        adjacency list, without information on the link weights.

    See Also
    --------
    LoadFromPajek : Loads a network from file in Pajek format.
    """
    # 0) SECURITY CHECK
    N = len(adjmatrix)
    if labels:
        if len(labels) != N:
            raise ValueError( "List of labels not aligned with network size" )

    # 1) OPEN THE TARGET FILE
    outfile = open(filepath, 'w')

    # 2) SAVE INFORMATION OF THE NODES
    print('*Vertices', N, file=outfile)

    # Wrie the list of nodes if labels have been given
    if labels:
        for i in range(N):
            # line = '%d "%s"' %(i+1, labels[i])
            line = '%d\t"%s"' % (i + 1, labels[i])
            print(line, file=outfile)

    # 3) SAVE THE LINKS
    # Save the links AND their WEIGHTS
    if weighted:
        # 3.1) Find whether weights are integers or floats
        if adjmatrix[0, 0].dtype in [np.uint8, np.uint, np.int8, np.int]:
            formatstring = '%d %d %d'
        elif adjmatrix[0, 0].dtype in [np.float16, np.float32, np.float, np.float64]:
            formatstring = '%d %d %f'

        # 3.2) Save the ARCS if directed
        if directed:
            print('*Arcs', file=outfile)
            for i in range(N):
                neighbours = adjmatrix[i].nonzero()[0]
                for j in neighbours:
                    line = formatstring % (i + 1, j + 1, adjmatrix[i, j])
                    print(line, file=outfile)

        # 3.2) Save the EDGES, if undirected
        else:
            print('*Edges', file=outfile)
            for i in range(N):
                neighbours = adjmatrix[i].nonzero()[0]
                for j in neighbours:
                    if j > i:
                        line = formatstring % (i + 1, j + 1, adjmatrix[i, j])
                        print(line, file=outfile)

    # Save ONLY the adjacency list
    else:
        formatstring = '%d %d'

        # 3.1) Save the ARCS if directed
        if directed:
            print('*Arcs', file=outfile)
            for i in range(N):
                neighbours = adjmatrix[i].nonzero()[0]
                for j in neighbours:
                    line = formatstring % (i + 1, j + 1)
                    print(line, file=outfile)

        # 3.1) Save the EDGES, if undirected
        else:
            print('*Edges', file=outfile)
            for i in range(N):
                neighbours = adjmatrix[i].nonzero()[0]
                for j in neighbours:
                    if j > i:
                        line = formatstring % (i + 1, j + 1)
                        print(line, file=outfile)

    # 4) CLOSE FILE AND FINISH
    outfile.close()

def LoadLabels(filepath):
    """Reads the labels of nodes from a text file.

    Parameters
    ----------
    filepath : string
        The path and filename in which the labels are stored.

    Returns
    -------
    data : list
        A list of strings with the label of each node.

    See Also
    --------
    SaveLabels : Saves the labels of nodes into a file
    """
    with open(filepath, 'r') as datafile:
        lines = [line.strip() for line in datafile.readlines()]
        # filter for empty lines
        return [line for line in lines if line]

def SaveLabels(filepath, labels):
    """Saves the labels of nodes into a text file.

    Parameters
    ----------
    labels : list-like
        A list or tuple of the labels of the nodes as strings.
    filepath : string
        The path and filename in which the labels will be stored.

    See Also
    --------
    LoadLabels : Reads the labels of nodes from a file
    """
    # 1) Create a string with all the text to be stored
    text = '\n'.join(labels)

    # 2) Open the datafile and save the text
    with open(filepath, 'w') as outfile:
        outfile.write(text)

def LoadPartition(filepath, sep=r'\s'):
    """Reads a partition of nodes from a text file.

    Every line of the file must contain the list of nodes in one community.

    Returns
    -------
    partition : list
        A list containing as many lists of nodes as communities.

    See Also
    --------
    SavePartition : Saves a partition of nodes into a file.
    """

    # 0) READ THE DATA FILE
    with open(filepath, 'r') as datafile:
        partition = []
        for line in datafile:
            # 1) CUT AT SEPARATOR
            newcom = re.split(sep, line.strip())
            # 2) CONVERT TO DECIMALS AND APPEND TO LIST
            partition.append([int(x) for x in newcom])

        return partition

def SavePartition(filepath, partition, sep=' '):
    """Saves a partition of nodes into a text file.

    Parameters
    ----------
    partition : list
        A list containing as many lists of nodes as communities.
    filepath : string
        The target file in which the partition is to be saved.
    labels : boolean, optional

    See Also
    --------
    LoadPartition : Reads a partition of nodes saved in a text file.
    """
    def _com_to_string(com):
        """ helper function to string format node weights """
        return sep.join([str(x) for x in com])

    # joining lines with line-breaks
    text = '\n'.join([_com_to_string(com) for com in partition])

    # writing formatted matrix to file
    with open(filepath, 'w') as outfile:
        outfile.write(text)

def ExtractSubmatrix(adjmatrix, nodelist1, nodelist2=[]):
    """Returns the sub-matrix composed by a set of nodes.

    Parameters
    ----------
    adjmatrix : ndarray of rank-2.
        The matrix from which the submatrix will be extracted.
    nodelist1 : arra_like. list, tuple, set or ndarray of rank-1.
        A list of indices with the rows of the matrix to be extracted.
        Prefered type is 'ndarray'. All other sequence types (lists, tuples and
        sets) will be converted to an ndarray of dtype = numpy.int.
    nodelist2 : arra_like. list, tuple, set or ndarray of rank-1 (optional)
        If not given (nodelist2=None), then nodelist2 is nodelist1.
        A list of indices with the columns of the matrix to be extracted.
        Prefered type is 'ndarray'. All other sequence types (lists, tuples and
        sets) will be converted to an ndarray of dtype = numpy.int.

    Returns
    -------
    A submatrix of 'adjmatrix' of size n1 x n2, where n1 and n2 are the
    lengths of nodelist1 and nodelist2 respectively.

    Examples
    --------
    >>> testmat = arange(1,17).reshape(4,4)
    >>> ExtractSubmatrix(testmat,[0,3])
    array([[ 1,  4],
           [13, 16]])
    >>>
    >>> ExtractSubmatrix(testmat,[0,3],[1,2,3])
    array([[ 2,  3,  4],
           [14, 15, 16]])

    """
    # 0) CHECK WHETHER LISTS OF NODES ARE GIVEN AS ARRAYS. OTHERWISE, CONVERT.
    # Check nodelist1
    if type(nodelist1) != np.ndarray:
        # warnings.simplefilter('once', UserWarning)
        # warnings.warn("Prefered type for parameter 'nodelist1' is numpy.ndarray. Lists, tuples and sets are allowed but are converted to ndarrays.", \
        #             stacklevel=2)

        if type(nodelist1) == set:
            nodelist1 = list(nodelist1)
        if type(nodelist1) in [list, tuple]:
            nodelist1 = np.array(nodelist1,np.int)

    # Check nodelist2
    if len(nodelist2) == 0:
        nodelist2 = nodelist1.copy()
    else:
        if type(nodelist2) != np.ndarray:
            # warnings.simplefilter('once', UserWarning)
            # warnings.warn("Prefered type for parameter 'nodelist2' is numpy.ndarray. Lists, tuples and sets are allowed but are converted to ndarrays.", \
            #                 stacklevel=2)

            if type(nodelist2) == set:
                nodelist2 = list(nodelist2)
            if type(nodelist2) in [list, tuple]:
                nodelist2 = np.array(nodelist2,np.int)

    # 1) CREATE LISTS OF INDICES FOR SLICING
    N1 = len(nodelist1)
    N2 = len(nodelist2)
    xindices = np.zeros(N1*N2, np.int)
    for ncounter, node in enumerate(nodelist1):
        startidx = ncounter*N2
        endidx = startidx + N2
        xindices[startidx:endidx] = node

    yindices = np.zeros(N1*N2, np.int)
    for n in range(N1):
        startidx = n*N2
        endidx = startidx + N2
        yindices[startidx:endidx] = nodelist2

    # 2) EXTRACT THE SUBMATRIX AND FINISH
    return adjmatrix[xindices, yindices].reshape(N1, N2)

def SymmetriseMatrix(adjmatrix):
    """Converts a directed network into undirected by averaging the weights,
    i.e., bij = 1/2 * (aij + aji).

    Parameters
    ----------
    adjmatrix : ndarray of rank-2
        The adjacency matrix of the network.

    Returns
    -------
    An adjacency matrix of the same shape, of dype=float, with values
    """

    if galib.metrics.Reciprocity(adjmatrix) == 1:
    # if Reciprocity(adjmatrix) == 1:
        return adjmatrix
    else:
        return 0.5 * (adjmatrix + adjmatrix.T)

def LaplacianMatrix(adjmatrix):
    """Returns the Laplacian matrix of a given network.

    The Laplacian matrix of a network is the matrix L = D - A, where A is
    the adjacency matrix and D is a diagonal matrix with elements Dii being
    the corresponding row sum of A.

    Parameters
    ----------
    adjmatrix : ndarray of rank-2
        The adjacency matrix of the network.

    Returns
    -------
    laplacianmatrix : ndarray of same shape and dtype as adjmatrix

    Warning
    -------
    If the network is directed, L should be computed out of the transposed
    adjacency matrix, otherwise it would represent the dynamics of the nodes
    driven by their outputs, not by their inputs. We recommend to introduce
    adjmatrix.T into the function.
    """
    if adjmatrix.dtype in [np.uint, np.uint0, np.uint8, np.uint16, np.uint32, np.uint64]:
        adjmatrix = adjmatrix.astype(int)
    N = len(adjmatrix)

    laplacianmatrix = np.identity(N, dtype=adjmatrix.dtype) * adjmatrix.sum(axis=1)
    laplacianmatrix -= adjmatrix

    return laplacianmatrix

def CleanPaths(pathlist):
    """Finds and removes in-place repeated, opposite paths from a list of paths.

    Parameters
    ----------
    pathlist : list
        A list containing paths as list or array-like.

    See Also
    --------
    PathsAllinOne : Computes several graph distance and path measures at once.
    CleanCycles : Removes repeated cycles from a list of cycles.

    Notes
    -----
    1. The function modifies the parameter 'pathlist' in-place and it does
    not return a value. On the contrary, the function CleanCycles() does
    return another list.
    2. The function is intended to be used with the output of PathsAllinOne(),
    and ONLY in the case of undirected graphs because PathsAllinOne() returns
    all paths twice, from i to j and from j to i. Do not apply CleanPaths()
    to the output from directed networks.

    Examples
    --------
    >>> paths = [[1,2,3],[3,2,1],[2,4,7,8],[8,7,4,2]]
    >>> CleanPaths(paths)
    >>> paths
    [[1, 2, 3], [2, 4, 7, 8]]
    """
    for path1 in pathlist:
        for path2 in pathlist[::-1]:
            if path2[::-1] == path1:
                pathlist.remove(path2)
                break

## ARRAY AND MATRIX COMPARISONS ###############################################
def HammingDistance(array1, array2, normed=False):
    """Computes the Hamming distance between two arrays of same shape.

    The Hamming distance counts the fraction of elements that are different
    in two arrays. It is 0 when all the two arrays are identical and 1 when
    the two arrays have no common elements.

    Note
    ----
    In case of binary data, it can be normed to account for the random
    probability of coincidences in two arrays with n_1,0 and n_2,0 number
    of zero entries and n_1,1 and n_2,1 number of ones.

    Parameters
    ----------
    array1 : ndarray of any shape and dimension
    array2 : ndarray of same shape and dimension as array1
    normed : Boolean, optional
        If True, the function computes the expected Hamming distance between
        the two vectors. The expected value is based on the expected number
        of coincidences that would be found if both arrays were random
        binary vectors with n_1 and n_2 number of ones respectively (giving
        length - n_1 and length - n_2 number of zeros).

    Returns
    -------
    h_dist : floating point.
        The Hamming distance between the two arrays.
    exp_hdist : floating point. Optional, only if 'expected = True'
        The expected Hamming distance between the vectors
    """
    # 0) PREPARE FOR CALCULATIONS
    # 0.1) Convert the arrays into rank-1 arrays
    if len(np.shape(array1)) > 1:
        array1 = array1.reshape(-1)
    if len(np.shape(array2)) > 1:
        array2 = array2.reshape(-1)

    # 0.2) Security check
    if len(array1) != len(array2):
        raise ValueError( "Arrays are not aligned" )

    # 1) COUNT THE NUMBER OF COINCIDENCES
    similarity = (array1 == array2)
    n_equal = similarity.sum()

    # 2) COMPUTE THE HAMMING DISTANCE
    length = len(array1)
    h_dist = 1. - float(n_equal) / length

    # 3) RETURN RESULT ACCORDING TO OPTIONS
    # Standard Hamming distance
    if not normed:
        return h_dist

    # Normalized Hamming distance
    else:
        # Count the number of ones in the two arrays
        n_1 = len(array1.nonzero()[0])
        n_2 = len(array2.nonzero()[0])

        # Estimate the expected number of random coincidences
        exp_nc = 1.0 / length * (n_1 * n_2 + (length - n_1) * (length - n_2))

        # The expected Hamming distance
        exp_hdist = 1.0 - exp_nc / float(length)

        return h_dist, exp_hdist

## SOME ADDITIONAL MATH ######################################################
def NonZeroMin(data):
    """Returns the smallest non-zero value in an array.

    Parameters
    ----------
    data : array-like
        A list, tuple or array of numbers.

    Returns
    -------
    An integer or real value, depending on data's dtype.
    """

    # 1) Convert lists and tuples into arrays
    if type(data) != 'numpy.ndarray':
        data = np.array(data)

    # 2) Find the minimal non-zero value and return.
    idx = np.where(data)
    return data[idx].min()

def CumulativeDistribution(data, nbins, range=None, normed=True, centerbins=False):
    """Computes the cumulative distribution of a dataset.

    Parameters
    ----------
    data : array_like
        Input data. The histogram is computed over the flattened array.
    nbins : int or float
        If `nbins` is an int, it defines the number of equal-width
        bins in the given range (10, by default). If `bins` is a sequence,
        it defines the bin edges, including the rightmost edge, allowing
    range : (float, float), optional
        The lower and upper range of the bins. If not provided, range
        is simply (data.min(), data.max()). Values outside the range are
        ignored.
    normed : bool, optional
        If False, the result will contain the number of samples
        in each bin.  If True, the result is the value of the
        probability *density* function at the bin, normalized such that
        the *integral* over the range is 1. Note that the sum of the
        histogram values will not be equal to 1 unless bins of unity
        width are chosen; it is not a probability *mass* function.
    centerbins : bool, optional
        If False, the positions of the bins ('xdata' variable) returned
        correspond to the lower limit of the each bin, as it is returned
        originally by the histogram() function of numpy. If True, 'xdata' will
        contain the value at the center of the bin.

    Returns
    -------
    xdata : ndarray
        Returns the bin location. This will be the left edge of the bins if
        option 'centerbins=False', and the center of the bin if option
        centerbins=True.
    ydata : ndarray
        The values of the cumulative probability.

    See Also
    --------
    ''histogram()'' function for further documentation.
    """

    # 1) COMPUTE THE DISTRIBUTION OF THE DATA
    ydata, xdata = np.histogram(data, nbins, range, normed)

    # 1.1) Compute the cumulative sum of the probability
    ydata = ydata.cumsum()

    # 2) RETURN THE RESULTS
    if centerbins:
        dif = 0.5 * (xdata[-1] - xdata[0]) / nbins
        xdata += dif

    if normed:
        norm = 1.0 / ydata[-1]
        ydata *= norm

        return xdata[:-1], ydata

    else:
        return xdata[:-1], ydata

def Factorial(x):
    """Computes the factorial of an integer number.
    """
    # 0) SECURITY CHECK
    if not isinstance(x, int):
        raise ValueError( "'Factorial' function only accepts integers" )

    # 1) COMPUTE THE FACTORIAL
    if x == 0 or x == 1:
        return 1
    else:
        return functools.reduce(lambda x, y: x * y, range(1, x + 1))

def BinomialCoefficient(n, m):
    """Computes the binomial coefficient of n over m.
    """
    if m == 0:
        return 1

    elif m == 1:
        return n

    else:
        ma = max(n - m, m)
        mi = min(n - m, m)

        enum = functools.reduce(lambda x, y: x * y, range(ma + 1, n + 1), 1)

        return enum / Factorial(mi)

def Quartiles(data):
    """
    Finds the 1st, 2nd and 3rd quartiles of a dataset.

    Parameters
    ----------
    data : array_like
        Input data.

    Returns
    -------
    Q1 : float. The first quartile.
    Q2 : float. The second quartile or median.
    Q3 : float. The third quartile.

    See Also
    ---------
    numpy.percentile
    """
    q = np.percentile(data, [25, 50, 75])

    return q[0], q[1], q[2]

def AllPermutations(data):
    """Given a set, it returns all possible permutations.

    Parameters
    ----------
    data : array_like
        A list, tuple or array containing data.

    Returns
    -------
    permutationlist : list
        A list containing all possible permutations of the elements in data.

    See Also
    --------
    itertools.permutations, AllCombinations, AllBipartitions

    Examples
    --------
    >>> AllPermutations((1,2,3))
    >>> [(1, 2, 3), (1, 3, 2), (2, 1, 3), (2, 3, 1), (3, 1, 2), (3, 2, 1)]

    >>> AllPermutation(('test','case',3))
    >>> [('test', 'case', 3),
        ('test', 3, 'case'),
        ('case', 'test', 3),
        ('case', 3, 'test'),
        (3, 'test', 'case'),
        (3, 'case', 'test')]
    """
    if len(data) <= 1:
        return data

    return [p for p in itertools.permutations(data)]

def AllCombinations(data, comblength):
    """Given a set, finds all combinations of given size.

    Returns a list with all possible permutations of length <= comblength
    of the elements in data. If N is the length of the data, the number
    of combinations of length 'comblength' is the binomial coefficient
    BinomialCoefficient(N,comblength).

    Parameters
    ----------
    data : array_like
        A list, tuple or array containing data.
    comblength : int
        Number of elements of the permutation lists.

    Returns
    -------
    combinationlist : list
        A list of all possible permutations of length 'comblength' with the
        elements in 'data'.

    See Also
    --------
    itertools.combinations, AllPermutations, AllBipartitions

    Examples
    --------
    >>>
    >>> [(1, 2), (1, 3), (1, 4), (2, 3), (2, 4), (3, 4)]
    >>>
    >>> AllCombinations((1,2,3,4), 3)
    >>> [(1, 2, 3), (1, 2, 4), (1, 3, 4), (2, 3, 4)]
    >>>
    >>> AllCombinations(('s','p','a','m'), 3)
    >>> [('s', 'p', 'a'), ('s', 'p', 'm'), ('s', 'a', 'm'), ('p', 'a', 'm')]
    """
    return [c for c in itertools.combinations(data, comblength)]

def AllBipartitions(data, comblength=None):
    """Given a set, finds all possilbe bipartitions.

    If optional 'comblength=None', it will return all bipartitions of
    all sizes. If 'comblength' is an integer, it will return all bipartitions
    in which one of the subsets has size N1 = comblength and complementary
    has size N2 = len(data) - comblength.

    Parameters
    ----------
    data : array_like
        A list, tuple or array containing data.
    comblength : integer, optional
        The size of one of the two complementary subsets.

    Returns
    -------
    result : list
        A list with all possible bipartitions of the elements in 'data'.

    See Also
    --------
    AllCombinations, AllPermutations

    Examples
    --------
    >>> AllBipartitions((1,2,3,4))
    >>> [((1,), (2, 3, 4)),
        ((2,), (1, 3, 4)),
        ((3,), (1, 2, 4)),
        ((4,), (1, 2, 3)),
        ((1, 2), (3, 4)),
        ((1, 3), (2, 4)),
        ((1, 4), (2, 3))]
    >>>
    >>> AllBipartitions(('s','p','a','m'))
    >>> [(('s',), ('a', 'p', 'm')),
        (('p',), ('a', 's', 'm')),
        (('a',), ('p', 's', 'm')),
        (('m',), ('a', 'p', 's')),
        (('s', 'p'), ('a', 'm')),
        (('s', 'a'), ('p', 'm')),
        (('s', 'm'), ('a', 'p'))]
    """
    # 0) PREPARE FOR CALCULATIONS
    Ndata = len(data)
    iseven = Ndata % 2 == 0
    Nmax = Ndata // 2
    setdata = set(data)

    # Security check, avoid confusing comblength=0 with None
    if comblength == 0:
        raise ValueError( "'comblength' out of range, please make sure 0 < comblength < len(data)." )

    ########################################################################
    # 1) DEFINE THE TWO FUNCTIONS THAT WILL DO THE ACTUAL JOB
    def AllBipartitionsAllSizes(data):
        """Given a dataset, it finds all bipartitions of all sizes.
        """
        # 1) FIND ALL THE COMBINATIONS UP TO LENGTH < len(data)/2
        bipartitions = []
        for n in range(1, Nmax):
            # 1.1) Find all combinations of given size out of dataset
            combinations = [c for c in itertools.combinations(data, n)]
            # 1.2) Sort and find complementary sets
            for comb in combinations:
                complementary = tuple(setdata - set(comb))
                bipartitions.append((comb, complementary))

        # 2) FIND AND SORT THE BIPARTITIONS OF SIZE Nmax
        # 2.1) Find all combinations of size Nmax
        combinations = [c for c in itertools.combinations(data, Nmax)]
        # 2.2) Sort and find complementary sets
        ncombs = len(combinations)
        # Ignore repeated combinations if both subsets are of size = Nmax
        if iseven: ncombs = ncombs // 2

        for i in range(ncombs):
            comb = combinations[i]
            combset = set(comb)
            complementary = setdata - combset
            bipartitions.append((comb, tuple(complementary)))

        return bipartitions

    def AllBipartitionsGivenSize(data, comblength):
        """Given a dataset, finds all bipartitions in which one of the
        subsets is of size 'comblength'. The size of the complementary
        subset is N - 'comblength', where N is the size of the data.
        """
        # 0) SECURITY CHECKS
        comblength = int(comblength)
        if (comblength < 1 or comblength >= Ndata):
            raise ValueError( "'comblength' out of range, please make sure 0 < comblength < len(data)." )

        # 1) FIND ALL THE COMBINATIONS OF SIZE 'comblength'
        combinations = [c for c in itertools.combinations(data, comblength)]

        # 2) SORT AND FIND THE COMPLEMENTARY SETS
        bipartitions = []

        # If size of set is even, and we are looking for bipartitions of size
        # Nmax, then we need to remove repeated combinations
        if iseven and comblength == Nmax:
            ncombs = len(combinations)
            for i in range(ncombs // 2):
                comb = combinations[i]
                combset = set(comb)
                complementary = setdata - combset
                bipartitions.append((comb, tuple(complementary)))

        # For any other case, bipartitions are found as usual
        else:
            for comb in combinations:
                complementary = tuple(setdata - set(comb))
                bipartitions.append((comb, complementary))

        return bipartitions

    ########################################################################
    # 2) DO THE JOB
    if comblength:
        result = AllBipartitionsGivenSize(data, comblength)
    else:
        result = AllBipartitionsAllSizes(data)

    return result

def MeanCorrelation(data, tolerance=10**(-15)):
    """Computes the Fisher-corrected mean value of correlation values.

    Parameters
    ----------
    data : array-like ndarray or list
        A tuple, list or array containing the values of correlation that
        are to be averaged. Accepts only 1-dimensional data, no matrices.
    tolerance : float
        Small jitter in the correlation values around exact 1.0 and -1.0
        that will also be accepted. Largest correlation values are
        set to 1.0 - tolerance and smallest correlation values are set to
        -1.0 + tolerance. This avoids +/- inf values to be returned by the
        arctanh() function. A single inf value would totally bias the mean.

    Returns
    -------
    float value
        Mean value of correlation in the range (-1,1).
    """
    # Security checks and data preparation
    dataset = np.array(data, np.float64)
    if len(np.shape(dataset)) > 1:
        dataset = dataset.flatten()

    if dataset.max() > 1.0 + tolerance:
        raise ValueError( "Correlation values larger than 1.0 in data" )
    if dataset.min() < -1.0 - tolerance:
        raise ValueError( "Correlation values smaller than -1.0 in data" )

    # Remove extremal values to avoid infinite values in the arctanh(x).
    idx = np.where(data >= 1.0)
    if idx:
        data[idx] = 1.0 - tolerance

    idx = np.where(data <= -1.0)
    if idx:
        data[idx] = -1.0 + tolerance

    # Compute the Fisher corrected
    newdata = np.arctanh(data)
    newdata = np.where(newdata==np.inf, 10,newdata)
    newdata = np.where(newdata==-np.inf, -10,newdata)

    return np.tanh(newdata.mean())

#



"""setup.py"""

from setuptools import setup, find_packages


with open("requirements.txt") as reqs_file:
    REQS = [line.rstrip() for line in reqs_file.readlines() if line[0] not in ['\n', '-', '#']]

config = {
    'name': 'galib',
    'description': ('A library for graph analysis in Python / NumPy.'),
    # Check these exist in PyPI
    'keywords': 'graph theory, complex networks, network analysis',
    'classifiers': [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    'author': 'Gorka Zamora-Lopez',
    'author_email': 'galib@Zamora-Lopez.xyz',
    'url': 'https://github.com/gorkazl/pyGAlib',
    'version': '1.0.2',
    'license': 'Apache License 2.0',
    'install_requires': REQS,
    'packages': find_packages(exclude=['doc', '*tests*']),
    'scripts': [],
    'include_package_data': True
    }

setup(**config)



# -*- coding: utf-8 -*-
# Copyright (c) 2013 - 2018, Gorka Zamora-López <gorka@Zamora-Lopez.xyz>
#
# Released under the Apache License, Version 2.0 (the "License");
# you may not use this software except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0

"""
Graph Analysis Library
======================

A package for the analysis of graphs and complex networks in Python.
Compatible with Python 2.7 and 3.X

GAlib treats networks as adjacency matrices, represented as 2D NumPy arrays,
thus taking advantage of the faster operations over pure Pythen data types.
The core of the library consists of three modules:

metrics
    Basic graph metrics (degrees, clustering, graph distance, etc)
models
    Generation of synthetic networks and randomization.
tools
    Miscelaneous helper functions.

The rest of modules provide additional functionalities:

metrics_numba
    Uses package Numba to accelerate calculation of some metrics.
models_numba
    Uses package Numba to accelerate generation of some graph models.
extra.py
    Additional measures and functionalities related to network analysis.

USING pyGAlib
^^^^^^^^^^^^^^
The library is organised in the following modules:

- *metrics.py*: Common graph metrics (degrees, clustering, graph distance, etc)
- *models.py*: Generation of synthetic networks and randomization.
- *tools.py*: Miscelaneous helper functions.
- *metrics_numba.py*: Uses the Numba package to accelerate calculation of some
metrics.
- *models_numba.py*: Uses the Numba package to accelerate generation of some
graph models.
- *extra.py*: Additional measures and functionalities related to network analysis.

Getting started
***************
Since pyGAlib depends on NumPy, it is recommended to import NumPy first.
Although this is not necessary for loading pyGAlib, NumPy functionalities and
array manipulation will be often needed. Try importing pyGAlib: ::

    >>> import numpy as np
    >>> import galib

..note::
    Importing galib imports also all functions in module *metrics.py*
    into its namespace. The rest of modules are imported separately. Therefore,
    if the import is relative those functions can be called as, e.g.,

::

    >>> import galib
    >>> ...
    >>> deg = galib.Degree(net)
    >>> C, Cnodes = galib.Clustering(net)

See that we did not have to call ``galib.metrics.Degree(net)``. In the case of
an absolute import (using an asterisk ``*``) all functions in *metrics.py* are
imported to the base namespace:  ::

    >>> from galib import *
    >>> ...
    >>> deg = Degree(net)
    >>> C, Cnodes = Clustering(net)

Example
*******
Let's generate a random graph following the Erdos-Renyi model, G(N,p), with
*N = 100* nodes and link probability *p = 0.1*:  ::

    >>> import galib
    >>> import galib.models
    >>> N = 100; p = 0.1
    >>> net = galib.models.ErdosRenyiGraph(N, p, directed=False)

Here, *net* is the adjacency matrix of the random graph represented as a
2-dimensional NumPy array. Let's calculate some basic properties.  ::

    >>> galib.Density(net)
    0.09838383838383838

As expected, the density of an Erdos-Renyi random graph is close to the
*p = 0.1* value given. We now calculate the degree of every node:  ::

    >>> deg = galib.Degree(net)
    >>> deg
    array([10,  7, 10, 10, 11,  7,  5, 11, 13, 12, 14, 13,  8, 10,  9,  8,  7,
       10, 11,  9, 11, 11,  8, 10,  5,  9, 13, 10, 13, 12, 12, 11, 11,  7,
       13, 11,  7, 10, 10,  6, 12, 10,  6, 10,  7,  6,  9, 10,  9,  9,  7,
        9,  8, 13, 10,  9,  7,  7, 11,  8, 13,  6,  7, 12, 14,  6,  5, 11,
        5, 12, 14, 14, 13,  8,  7, 12,  4, 19,  9, 13,  7, 10, 15, 15,  4,
        9,  7, 12,  7,  8, 12,  4, 11, 12,  6, 13,  6, 12, 16, 12])

The degree is returned as a numpy array of rank 1, of integer type. The function
``Clustering()`` returns both the global clustering coefficient and the local
clustering of every node:  ::

    >>> C, Cnodes = galib.Clustering(net)
    >>> C
    0.096051227321238
    >>> Cnodes
    array([0.13333333, 0.04761905, 0.04444444, 0.08888889, 0.10909091,
       0.19047619, 0.3       , 0.12727273, 0.08974359, 0.06060606,
      	... ... ...
      	... ... ...
       0.04545455, 0.        , 0.10909091, 0.13636364, 0.        ,
       0.07692308, 0.13333333, 0.09090909, 0.08333333, 0.10606061])

We compute the pair-wise graph distance between all the nodes using the
Floyd-Warshall algorithm, which is returned as a matrix ``dij`` (numpy array of
rank 2):  ::

	>>> dij = galib.FloydWarshall(net)
	>>> avlen = (dij.sum() - dij.trace()) / (N*(N-1))
	>>> avlen
	2.248080808080808

.. note::
    For calculating the distance matrix of larger networks please use
    the version of the function located in the module *metrics_numba.py*. Here,
    the function ``FloydWarshall_Numba()`` works the same but makes use of the
    Numba library to significantly speed the calculation.

Most network generators and graph metrics in pyGAlib work with directed graphs
as well. Check for the optional parameter ``directed``. Following the example
above, we generate a directed Erdos-Renyi graph and calculate its input and
output degrees for every node:  ::

    >>> net = galib.models.ErdosRenyiGraph(N, p, directed=True)
    >>> galib.Density(net)
    0.10272727272727272
    >>> indeg, outdeg = galib.Degree(net, directed=True)
    >>> indeg
    array([17,  7,  9,  8, 11, 10,  9,  8, 13, 13,  5,  9, 13,  9, 10, 10, 13,
       10,  9,  9,  7, 11, 13, 10,  4, 15, 11, 11, 10,  6,  6,  8,  8,  8,
       11,  8,  4, 12,  8, 13, 13, 14, 12,  5,  6,  5, 16, 12,  5, 10,  9,
       13,  8,  9,  7,  8, 13, 14,  9, 18,  7, 11,  5,  4, 12,  8,  8, 10,
        7,  9, 15, 12, 14,  9, 15, 11, 13, 12, 15, 10, 11, 11, 15,  7, 10,
       13,  7, 14,  9, 16, 11, 11,  6, 18,  7,  4, 14, 12, 12, 10])
    >>> outdeg
    array([ 9, 10,  7,  9, 12,  9, 19,  9, 11, 16, 11, 12, 11, 15, 11,  6,  9,
        8, 11, 12,  9, 13,  9,  8, 11,  6,  7, 11, 11, 12, 10,  8, 11, 12,
       10, 12, 13,  8, 18, 11,  8, 13, 10,  8, 10, 10, 11,  8, 11, 11, 11,
       10, 11, 10,  9, 12,  6, 10,  7, 10, 10, 11, 15, 12, 11,  7, 10,  8,
        5, 11,  7, 11, 13,  8,  5,  6, 13, 11, 10, 13,  7,  6, 13, 11,  8,
        8, 10,  6, 10,  9, 12, 15, 11,  9, 15, 11,  7,  8, 11, 10])

Data I/O
********
Since GAlib is based on NumPy arrays, saving and reading of adjacency matrices,
as well as any other output of GAlib functions, can be performed using the usual
data I/O functionalities of NumPy. See for example the documentation for
functions: ``loadtxt()``, ``savetxt()``, ``load()``, ``save()`` and ``savez()``.
The *tools.py* module in pyGAlib provides also some data conversion
functionalities.

How to find further documentation
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
While working in an interactive session, after importing a module, the built-in
``help()`` function will show further details:  ::

    >>> help(modulename)

The help for galib (``help(galib)``) shows the general summary of the package
and a list of all the modules in the library. The help for each module,
e.g., ``help(galib.metrics)`` or ``help(galib.models)`` will display module
specific information and a list of all the functions in the module. For further
details regarding each function, type:  ::

    >>> help(galib.modulename.functionname)

For IPython and Jupyter notebook users the help command is replaced by a
question mark after the module's or function's name, e.g.:  ::

    >>> modulename?
    >>> functionname?

For questions, bug reports, etc, please write to <galib@Zamora-Lopez.xyz>, or
open an issue in GitHub.

License
-------

Copyright (c) 2013 - 2018, Gorka Zamora-López <gorka@Zamora-Lopez.xyz>

Released under the Apache License, Version 2.0 (the "License");
you may not use this software except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

or see the LICENSE.txt file.

.. note::
    Please, use the logos provided in the Branding/ folder whenever possible.

"""
from __future__ import absolute_import

from . import metrics
from .metrics import*


__author__ = "Gorka Zamora-Lopez"
__email__ = "galib@Zamora-Lopez.xyz"
__copyright__ = "Copyright 2013-2018"
__license__ = "Apache License version 2.0"
__version__ = "1.0.2"
__update__="05/05/2019"



#



# -*- coding: utf-8 -*-
# Copyright (c) 2013 - 2018, Gorka Zamora-López <gorka@Zamora-Lopez.xyz>
#
# Released under the Apache License, Version 2.0 (the "License");
# you may not use this software except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0

"""
==========================
ADDITIONAL FUNCTIONALITIES
==========================

This module hosts functionalities and measures related to the study of complex
networks which do not fit in the more strict classification of graph metrics,
models and tools. These correspond to functionalities I have developed and
published during my research, which might be of interest for some users or
readers. It also contains adaptation into Python of measures published by
third-party authors which I needed for my own research.

ESTIMATION OF EXPECTED CROSS-CORRELATION
========================================
TopologicalSimilarity
    Computes the expected cross-correlation matrix of a given network.
ExponentialMapping
    Computes the expected cross-correlation matrix of a given network.
CovarianceLinearGaussian
    Approximated covariance matrix of a system of Gaussian noise sources.

NETWORK (DYNAMICS) COMPLEXITY MEASURES
=======================================
FunctionalComplexity
    Calculates the functional complexity of a correlation(-like) matrix.
NeuralComplexity
    Calculates the neural complexity of a correlation(-like) matrix.
NeuralComplexity_Sampled
    Calculates the neural complexity of a correlation(-like) matrix (Faster).


...moduleauthor:: Gorka Zamora-Lopez <galib@zamora-lopez.xyz>

"""
from __future__ import division, print_function, absolute_import

import numpy as np
import numpy.random
import scipy.linalg

from . import tools


############################################################################
"""ESTIMATION OF EXPECTED CROSS-CORRELATION"""
def TopologicalSimilarity(adjmatrix, coupling):
    """Computes the expected cross-correlation matrix of a given network.

    This function estimates the expected correlations between its nodes that
    a dynamical process on the network tends to generate. The estimation assumes
    that the "influence" of one node over another distributes over all possible
    paths but that the influence decays along the paths. This decay is estimated
    as the communicability matrix. Pairwise correlation is assumed, according
    to the patterns of inputs two nodes tend to receive.
    See further details in our paper:
    "Bettinardi et al. Chaos 27:047409 (2017)"

    Parameters
    ----------
    adjmatrix : ndarray of rank-2
        The adjacency matrix of the network.
    coupling : real valued number
        The coupling strength of all connections.

    Returns
    -------
    corrmatrix : ndarray of rank-2, same size as adjmatrix.
        The correlation matrix of the expected functional connectivity.

    See Also
    --------
    ExponentialMapping : Expected cross-correlation matrix of a given network.
    CovarianceLinearGaussian : Covariance matrix of an Ornstein-Uhlenbeck process.

    Citation
    --------
    R.G. Bettinardi, G. Deco et al. "How structure sculpts function: Unveiling
    the contribution of anatomical connectivity to the brain’s spontaneous
    correlation structure" Chaos 27, 047409 (2017).
    """
    # 0) Security check on the dtype
    if adjmatrix.dtype not in ['float32','float64','float']:
        adjmatrix = adjmatrix.astype(np.float64)

    # 1) Compute the communicability matrix
    Cmatrix = scipy.linalg.expm(coupling*adjmatrix).T

    # 2) Compute the product between all columns (Cmatrix was transposed)
    #    and the normalization
    corrmatrix = np.inner(Cmatrix, Cmatrix)
    norms = scipy.linalg.norm(Cmatrix, axis=1)
    normmatrix = np.outer(norms, norms.T)

    corrmatrix = corrmatrix / normmatrix
    return corrmatrix

def ExponentialMapping(adjmatrix, coupling, partialcorr=False):
    """Computes the expected cross-correlation matrix of a given network.

    This function estimates the expected cross-correlation matrix of a dynamical
    process on a network, assuming that each node is a linear Gaussian noise
    source and that the "influence" of one node over another distributes over
    all possible paths but that the influence decays along the paths. This decay
    is estimated as the communicability matrix. See further details in our paper:
    "Zamora-Lopez et al. Sci. Reps. 6:38424 (2016)."

    NOTE: The output of the Exponential Mapping is numerically the same as the
    output of the Topological Similarity measure, when the noise level for all
    nodes is the same, and when the decay rate for all nodes is the same, as
    shown in "Bettinardi et al. Chaos 27:047409 (2017) (Supp. Material)."
    I recommend to use TopologicalSimilarity() instead. I added this function
    here only for reference from interested readers of the paper above.

    Parameters
    ----------
    adjmatrix : ndarray of rank-2
        The adjacency matrix of the network.
    coupling : real valued number
        The coupling strength of all connections.
    partialcorr : boolean (optional)
        If 'True', the function returs the partial correlation matrix.

    Returns
    -------
    corrmatrix : ndarray of rank-2, same size as adjmatrix.
        The correlation matrix of the expected functional connectivity.

    See Also
    --------
    TopologicalSimilarity : Expected cross-correlation matrix of a given network.
    CovarianceLinearGaussian : Covariance matrix of an Ornstein-Uhlenbeck process.

    Citation
    --------
    G. Zamora-Lopez, Y. Chen, G. Deco, M.L. Kringelbach & C.S. Zhou,
    "Functional complexity emerging from anatomical constraints in the brain:
    the significance of network modularity and Rich-clubs." Scientific Reports,
    6:38424 (2016).
    """
    # 0) Security check on the dtype
    if adjmatrix.dtype not in ['float32','float64','float']:
        adjmatrix = adjmatrix.astype(np.float64)

    # 1) Compute the node-to-node influence
    Qmatrix = scipy.linalg.expm(coupling*adjmatrix)
    # 2) Normalize the Qmatrix to avoid numerical overload
    maxQvalue = max( abs(Qmatrix.min()), abs(Qmatrix.max()) )

    if partialcorr:
        #Qmatrix = Qmatrix / maxQvalue
        # 3) Compute the covariance matrix and its inverse
        covmatrix = np.dot(Qmatrix.T, Qmatrix)
        covmatrix = scipy.linalg.inv(covmatrix)
        # 4) Convert to a correlation matrix and return
        diagvalues = np.diag(covmatrix)
        normmatrix = np.sqrt( np.outer(diagvalues,diagvalues) )
        corrmatrix = - covmatrix / normmatrix

    else:
        Qmatrix = Qmatrix / maxQvalue
        # 3) Compute the covariance matrix
        covmatrix = np.dot(Qmatrix.T, Qmatrix)
        # 4) Convert to a correlation matrix and return
        diagvalues = np.diag(covmatrix)
        normmatrix = np.sqrt( np.outer(diagvalues,diagvalues) )
        corrmatrix = covmatrix / normmatrix

    return corrmatrix

def CovarianceLinearGaussian(adjmatrix, coupling, noiselevel=1.0, noisearray=[]):
    """Estimates the covariance matrix of a system of Gaussian noise sources.

    This function returns an analytical approximation of the time-average
    covariance matrix of a system of linearly coupled Gaussian noise sources.
    Each node of the system (network) is described as fixed point which receives
    a Gaussian noise and the input from its neighbours. Given that X is the
    vector of states of the nodes and A is the adjacency matrix of the
    connectivity and 'g' a global coupling strength applied to all links, the
    equation of motion of the system is approximately:
        Xdot = -a*X + g* A*X + D*noise
    where 'a' is the decay rate parameter and D is the intensity of the Gaussian
    white noise of variance = 1.
    This model has received many different names. It is often refered as the
    Ornstein-Uhlenbeck process, the continous version of a discrete-time
    autorregressive model.

    The analytical approximation in this function is coded to match the
    description given in:
        "Tononi et al. PNAS 91, 5033-5037 (1994)"
    which has been often used in the neuroscience / brain connectivity
    literature.

    - NOTE 1: Because this model diverges when g equals the largest eigenvalue
    of the adjacency matrix, it is highly recommended to provide as input the
    normalised version of the adjacency matrix. That is, the adjacency
    matrix divided by its spectral diameter. In that case, 'g' takes values
    between 0 and 1. See "Zamora-Lopez et al. Front. Neuroinform. 4, (2010)".

    - NOTE 2: In case the network is directed, the covariance matrix is
    computed out of the transposed adjacency matrix, such that the dynamics of
    a node is governed by the inputs i receives, instead of by its outputs.

    Parameters
    ----------
    adjmatrix : ndarray of rank-2
        The adjacency matrix of the network. It can be either weighted or
        binary, directed or undirected. It is recommended to provide the
        adjacency matrix normalised by its largest eigenvalue (or spectral
        diameter).
    coupling : real valued number
        The coupling strength of all connections. 'coupling' should be smaller
        than the spectral diameter of 'adjmatrix'. If this has been already
        normalised, then g can take values between 0 and 1.
    noiselevel : real valued number (optional)
        The intensity of the noise, if all nodes should receive the same.
    noisearray : list, tuple or array_like (optional)
        An array contaning the individual noise level each node receives.
        If left empy, all nodes will receive the same noise level given by
        parameter 'noiselevel'. Else, 'noiselevel' will be replaced by the
        'noisearray' parameter. The length of 'noisearray' must be the number
        of nodes in the network.

    Returns
    -------
    COVmatrix : ndarray of rank-2
        The approximated time-average covariance matrix of the system.

    See Also
    --------
    TopologicalSimilarity : Expected cross-correlation matrix of a given network.
    ExponentialMapping : Expected cross-correlation matrix of a given network.
    """
    # 0) Security checks and preparations
    if adjmatrix.dtype not in ['float32', 'float64']:
        adjmatrix = adjmatrix.astype(np.float64)

    # 1) Construct the uncorrelated noise vector (diagonal matrix)
    N = len(adjmatrix)
    if len(noisearray) > 0:
        if len(noisearray) != N:
            raise ValueError( "'noisearray' not aligned with 'adjmatrix'" )
        R = np.array(noisearray, np.float64) * np.identity(N).astype(np.float64)
    else:
        R = noiselevel * np.identity(N).astype(np.float64)

    # 2.2) Compute the Q matrix
    Qmatrix = scipy.linalg.inv(np.identity(N,np.float64) - coupling*adjmatrix)
    Qt = Qmatrix.T

    # 2.3) Compute the covariance matrix
    COVmatrix = np.dot(Qt,R**2)
    COVmatrix = np.dot(COVmatrix,Qmatrix)

    return COVmatrix

def Covmat2Corrmat(covmatrix):
    """Normalises a covariance matrix into a cross-correlation matrix.

    Parameters
    ----------
    covmatrix : ndarray of rank-2
        A covariance matrix.

    Returns
    -------
    corrmatrix : ndarray of rank-2
        The cross-correlation matrix.
    """
    # 0) Prepare for calculation
    if covmatrix.dtype not in ['float32','float64','float']:
        covmatrix = covmatrix.astype(np.float)

    # 1) Calculate the normalization factor for each pair
    diagvalues = np.diag(covmatrix).astype(np.float)
    normmatrix = np.sqrt(np.outer(diagvalues,diagvalues))

    # 2) Return result
    corrmatrix = covmatrix / normmatrix
    return corrmatrix


############################################################################
"""NETWORK (DYNAMICS) COMPLEXITY MEASURES"""
def FunctionalComplexity(corrmatrix, nbins=50, datarange=[0,1]):
    """Calculates the functional complexity of a correlation(-like) matrix.

    Given the cross-correlation matrix out of some network dynamics or an
    arbitrary multivariate dynamical system, this function computes the
    functional complexity of the system. Functional complexity takes value 0 if
    the elements or nodes are (dynamically or statistically) independent of
    each other. It takes value 1 when the elements or nodes are all fully
    correlated with each other. See definition, details and comparison
    to other measures in our paper:
    "Zamora-Lopez et al. Sci. Reps. 6:38424 (2016)."

    NOTE: Functional complexity can be estimated out of any matrix of pair-wise
    statistical associations between the elements of a network or multi-
    variate (dynamical or stochastic) system, e.g. the matrix of pairwise
    mutual information values. The range of accepted values may have to be
    adapted.

    Parameters
    ----------
    corrmatrix : ndarray of rank-2.
        The correlation or pair-wise statistical association matrix.
    nbins : integer (optional)
        Number of bins for which the distribution of values in the matrix shall
        be estimated.
    datarange : list, tuple or array_like (optional)
        A sequence of length = 2 containing the smallest and the largest values
        expected in the statistical association matrix.

    Returns
    -------
    fcomplexity : real valued number
        The functional complexity, ranging from 0 to 1.

    See Also
    --------
    NeuralComplexity : Measure of complexity by Tononi, Sporns and Edelmann.
    NeuralComplexity_Sampled : Measure of complexity by Tononi, Sporns et al.

    Citation
    --------
    G. Zamora-Lopez, Y. Chen, G. Deco, M.L. Kringelbach & C.S. Zhou,
    "Functional complexity emerging from anatomical constraints in the brain:
    the significance of network modularity and Rich-clubs." Scientific Reports,
    6:38424 (2016).

    """
    # 0) Security checks
    if len(np.shape(corrmatrix)) != 2:
        raise ValueError('Input data not a correlation matrix. Data not alligned.')
    if corrmatrix.min() < datarange[0]:
        raise ValueError('Input data not in range. Values smaller than range found.')
    if corrmatrix.max() > datarange[1]:
        raise ValueError('Input data not in range. Values larger than range found.')

    # 1) Use only the upper triangular values of the correlation matrix
    N = len(corrmatrix)
    idx = np.triu_indices(N,1)
    corrvalues = corrmatrix[idx]

    # 2) Compute the distribution of the correlation values
    # Convert corrvalues to 'float32' to avoid numerical issues. Don't ask me why.
    ydata, xdata = np.histogram(corrvalues.astype(np.float32), nbins, datarange, density=True)
    # Normalize probability. histogram() does not properly normalize.
    ydata /= ydata.sum()

    # 3) Compute functional complexity
    normfactor = 0.5 * float(nbins) / (nbins-1)
    uniformdistrib = 1./nbins * np.ones(nbins, np.float)
    fcomplexity = 1. - normfactor * np.add.reduce(abs(ydata - uniformdistrib))
    # fcomplexity = 1. - normfactor * (abs(ydata - uniformdistrib)).sum()

    # Clean some trash
    del xdata, corrvalues, uniformdistrib

    return fcomplexity

def NeuralComplexity(corrmatrix, bipartitions=None):
    """Calculates the neural complexity of a correlation(-like) matrix.

    This function calculates the 'neural complexity' measure proposed in
        "Tononi et al. PNAS 91, 5033-5037 (1994)"

    Parameters
    ----------
    corrmatrix : ndarray of rank-2.
        The correlation or pair-wise statistical association matrix.
    bipartitions : list
        A list of bipartitions of the nodes, e.g., [(1,2,3,4), (5,6)].

    Returns
    -------
    ncomplexity : real valued number
        The neural complexity value.

    See Also
    --------
    FunctionalComplexity : Functional complexity of a correlation(-like) matrix
    """
    # 0) Get ready for the calculations
    N = len(corrmatrix)

    # Find all possible bipartitions in a set of N nodes
    if not bipartitions:
        bipartitions = tools.AllBipartitions(np.arange(N))

    nmax = N // 2
    detcorrmat = scipy.linalg.det(corrmatrix)

    # 1) Calculate the mutual information, for all bipartitions of all sizes
    counter = 0
    avmibips = np.zeros(nmax+1, np.float64)
    for n in range(1,nmax+1):
        # Number of bipartitions for set of size n and (N-n)
        nbips = tools.Factorial(N) / (tools.Factorial(n) * tools.Factorial(N-n))

        # Correct nbips if N is even and if nbips = nmax
        if n == nmax and 2*n == N:
            nbips /= 2

        # Calculate mutual information for all bipartitions of sizes n, (N-n)
        nbips = int(nbips)
        mibips = np.zeros(nbips, np.float64)
        for i in range(nbips):
            # Choose a bipartition
            set1, set2 = bipartitions[counter+i]

            # Extract the submatrices for each bipartition
            submat1 = tools.ExtractSubmatrix(corrmatrix, set1)
            submat2 = tools.ExtractSubmatrix(corrmatrix, set2)

            # Compute the determinant of the submatrices
            detsubmat1 = scipy.linalg.det(submat1)
            detsubmat2 = scipy.linalg.det(submat2)

            # Compute the mutual information of the bipartition
            mibips[i] = 0.5 * ( np.log(detsubmat1) + np.log(detsubmat2) - np.log(detcorrmat) )

        counter += nbips

        # Average MI for all bipartitions of the same size
        avmibips[n] = mibips.mean()

    ncomplexity = avmibips.sum()
    return ncomplexity

def NeuralComplexity_Sampled(corrmatrix, maxiter=1000):
    """Calculates the neural complexity of a correlation(-like) matrix.

    This function calculates the 'neural complexity' measure proposed in
        "Tononi et al. PNAS 91, 5033-5037 (1994)"
    The difference with function 'NeuralComplexity()' is that the latter
    exhaustively computes the mutual information for all existing bipartitions
    of nodes in the network, which is computationally prohibitive for networks
    larger than N = 30 nodes. In this case, the calculation is limited to a
    a limited number of bipartitions (maxiter) among all possible. The
    sample bipartitions are generated at random.

    Parameters
    ----------
    corrmatrix : ndarray of rank-2.
        The correlation or pair-wise statistical association matrix.
    maxiter : integer
        Maximum number of bipartitions to generate for each size n, (N-n).

    Returns
    -------
    ncomplexity : real valued number
        The neural complexity value.

    See Also
    --------
    FunctionalComplexity : Functional complexity of a correlation(-like) matrix
    """
    # 0) GET READY FOR THE CALCULATIONS
    N = len(corrmatrix)

    nmax = N // 2
    detcorrmat = scipy.linalg.det(corrmatrix)
    factorialN = tools.Factorial(N)
    nodelist = np.arange(N)

    # 1) Calculate the mutual information, for all bipartitions of all sizes
    avmibips = np.zeros(nmax+1, np.float64)
    for n in range(1,nmax+1):
        nbips = factorialN / (tools.Factorial(n) * tools.Factorial(N-n))

        # Correct nbips is N is even and if nbips = nmax
        if n == nmax and 2*n == N:
            nbips /= 2

        # Average of all bipartitions of size n, if nbips is small
        if nbips <= maxiter:
            nbips = int(nbips)
            mibips = np.zeros(nbips, np.float64)
            # Create all the bipartitions of size n
            bipartitions = tools.AllBipartitions(nodelist, n)
            for i in range(nbips):
                # Choose a bipartition
                set1, set2 = bipartitions[i]

                # Extract the submatrices for each bipartition
                submat1 = tools.ExtractSubmatrix(corrmatrix, set1)
                submat2 = tools.ExtractSubmatrix(corrmatrix, set2)

                # Compute the determinant of the submatrices
                detsubmat1 = scipy.linalg.det(submat1)
                detsubmat2 = scipy.linalg.det(submat2)

                # Compute the mutual information of the bipartition
                mibips[i] = 0.5 * ( np.log(detsubmat1) + np.log(detsubmat2) - np.log(detcorrmat) )

        # If nbips is large, the average of randomly sampled bipartitions
        else:
            # print 'Sampling... n:', n
            mibips = np.zeros(maxiter, np.float64)
            for i in range(maxiter):
                # Generate a random bipartition of sizes n and N-n
                numpy.random.shuffle(nodelist)
                set1 = nodelist[:n]
                set2 = nodelist[n:]

                # Extract the submatrices for each bipartition
                submat1 = tools.ExtractSubmatrix(corrmatrix, set1)
                submat2 = tools.ExtractSubmatrix(corrmatrix, set2)

                # Compute the determinant of the submatrices
                detsubmat1 = scipy.linalg.det(submat1)
                detsubmat2 = scipy.linalg.det(submat2)

                # Compute the mutual information of the bipartition
                mibips[i] = 0.5 * ( np.log(detsubmat1) + np.log(detsubmat2) - np.log(detcorrmat) )

        # Average MI for all bipartitions of the same size
        avmibips[n] = mibips.mean()

    ncomplexity = avmibips.sum()
    return ncomplexity


#
